# 
# Generated by the nebius.base.protos.compiler.  DO NOT EDIT!
# 

"""Auto-generated Nebius SDK package for ``nebius.common.v1alpha1``"""

import builtins as builtins
import collections.abc as abc
import datetime as datetime
import google.protobuf.any_pb2 as any_pb2
import google.protobuf.descriptor as descriptor_1
import google.protobuf.message as message_1
import google.protobuf.timestamp_pb2 as timestamp_pb2
import google.rpc.status_pb2 as status_pb2
import nebius.aio.client as client
import nebius.aio.operation as operation
import nebius.aio.request as request_1
import nebius.aio.request_kwargs as request_kwargs
import nebius.aio.request_status as request_status
import nebius.api.nebius.common.v1alpha1.operation_pb2 as operation_pb2
import nebius.api.nebius.common.v1alpha1.operation_service_pb2 as operation_service_pb2
import nebius.base.protos.descriptor as descriptor
import nebius.base.protos.pb_classes as pb_classes
import nebius.base.protos.unset as unset
import nebius.base.protos.well_known as well_known_1
import typing_extensions as typing_extensions
#@ local imports here @#

# file: nebius/common/v1alpha1/operation.proto
class Operation(pb_classes.Message):
    __PB2_CLASS__ = operation_pb2.Operation
    __PB2_DESCRIPTOR__ = descriptor.DescriptorWrap[descriptor_1.Descriptor](".nebius.common.v1alpha1.Operation",operation_pb2.DESCRIPTOR,descriptor_1.Descriptor)
    __mask_functions__ = {
        "created_at": well_known_1.ts_mask,
        "finished_at": well_known_1.ts_mask,
        "status": well_known_1.status_mask,
    }
    
    class request_header(pb_classes.Message):
        """
        Request header is a container for all the values of a particular header of a request
        as there is no such thing as map<string, repeated string>
        """
        
        __PB2_CLASS__ = operation_pb2.Operation.request_header
        __PB2_DESCRIPTOR__ = descriptor.DescriptorWrap[descriptor_1.Descriptor](".nebius.common.v1alpha1.Operation.request_header",operation_pb2.DESCRIPTOR,descriptor_1.Descriptor)
        __mask_functions__ = {
        }
        
        def __init__(
            self,
            initial_message: message_1.Message|None = None,
            *,
            values: "abc.Iterable[builtins.str]|None|unset.UnsetType" = unset.Unset,
        ) -> None:
            super().__init__(initial_message)
            if not isinstance(values, unset.UnsetType):
                self.values = values
        
        def __dir__(self) ->abc.Iterable[builtins.str]:
            return [
                "values",
            ]
        
        @builtins.property
        def values(self) -> "abc.MutableSequence[builtins.str]":
            """
            The values of a particular header from a request
            """
            
            return super()._get_field("values", explicit_presence=False,
            wrap=pb_classes.Repeated,
            )
        @values.setter
        def values(self, value: "abc.Iterable[builtins.str]|None") -> None:
            return super()._set_field("values",value,explicit_presence=False,
            )
        
        __PY_TO_PB2__: builtins.dict[builtins.str,builtins.str] = {
            "values":"values",
        }
        
    
    class RequestHeadersEntry(pb_classes.Message):
        __PB2_CLASS__ = operation_pb2.Operation.RequestHeadersEntry
        __PB2_DESCRIPTOR__ = descriptor.DescriptorWrap[descriptor_1.Descriptor](".nebius.common.v1alpha1.Operation.RequestHeadersEntry",operation_pb2.DESCRIPTOR,descriptor_1.Descriptor)
        __mask_functions__ = {
        }
        
        def __init__(
            self,
            initial_message: message_1.Message|None = None,
            *,
            key: "builtins.str|None|unset.UnsetType" = unset.Unset,
            value: "Operation.request_header|operation_pb2.Operation.request_header|None|unset.UnsetType" = unset.Unset,
        ) -> None:
            super().__init__(initial_message)
            if not isinstance(key, unset.UnsetType):
                self.key = key
            if not isinstance(value, unset.UnsetType):
                self.value = value
        
        def __dir__(self) ->abc.Iterable[builtins.str]:
            return [
                "key",
                "value",
            ]
        
        @builtins.property
        def key(self) -> "builtins.str":
            return super()._get_field("key", explicit_presence=False,
            )
        @key.setter
        def key(self, value: "builtins.str|None") -> None:
            return super()._set_field("key",value,explicit_presence=False,
            )
        
        @builtins.property
        def value(self) -> "Operation.request_header":
            return super()._get_field("value", explicit_presence=False,
            wrap=Operation.request_header,
            )
        @value.setter
        def value(self, value: "Operation.request_header|operation_pb2.Operation.request_header|None") -> None:
            return super()._set_field("value",value,explicit_presence=False,
            )
        
        __PY_TO_PB2__: builtins.dict[builtins.str,builtins.str] = {
            "key":"key",
            "value":"value",
        }
        
    
    def __init__(
        self,
        initial_message: message_1.Message|None = None,
        *,
        id: "builtins.str|None|unset.UnsetType" = unset.Unset,
        description: "builtins.str|None|unset.UnsetType" = unset.Unset,
        created_at: "timestamp_pb2.Timestamp|datetime.datetime|None|unset.UnsetType" = unset.Unset,
        created_by: "builtins.str|None|unset.UnsetType" = unset.Unset,
        finished_at: "timestamp_pb2.Timestamp|datetime.datetime|None|unset.UnsetType" = unset.Unset,
        request: "any_pb2.Any|None|unset.UnsetType" = unset.Unset,
        request_headers: "abc.Mapping[builtins.str,Operation.request_header]|None|unset.UnsetType" = unset.Unset,
        resource_id: "builtins.str|None|unset.UnsetType" = unset.Unset,
        resource: "any_pb2.Any|None|unset.UnsetType" = unset.Unset,
        progress_data: "any_pb2.Any|None|unset.UnsetType" = unset.Unset,
        status: "status_pb2.Status|request_status.RequestStatus|None|unset.UnsetType" = unset.Unset,
    ) -> None:
        super().__init__(initial_message)
        if not isinstance(id, unset.UnsetType):
            self.id = id
        if not isinstance(description, unset.UnsetType):
            self.description = description
        if not isinstance(created_at, unset.UnsetType):
            self.created_at = created_at
        if not isinstance(created_by, unset.UnsetType):
            self.created_by = created_by
        if not isinstance(finished_at, unset.UnsetType):
            self.finished_at = finished_at
        if not isinstance(request, unset.UnsetType):
            self.request = request
        if not isinstance(request_headers, unset.UnsetType):
            self.request_headers = request_headers
        if not isinstance(resource_id, unset.UnsetType):
            self.resource_id = resource_id
        if not isinstance(resource, unset.UnsetType):
            self.resource = resource
        if not isinstance(progress_data, unset.UnsetType):
            self.progress_data = progress_data
        if not isinstance(status, unset.UnsetType):
            self.status = status
    
    def __dir__(self) ->abc.Iterable[builtins.str]:
        return [
            "id",
            "description",
            "created_at",
            "created_by",
            "finished_at",
            "request",
            "request_headers",
            "resource_id",
            "resource",
            "progress_data",
            "status",
            "request_header",
            "RequestHeadersEntry",
        ]
    
    @builtins.property
    def id(self) -> "builtins.str":
        """
        ID of the operation.
        """
        
        return super()._get_field("id", explicit_presence=False,
        )
    @id.setter
    def id(self, value: "builtins.str|None") -> None:
        return super()._set_field("id",value,explicit_presence=False,
        )
    
    @builtins.property
    def description(self) -> "builtins.str":
        """
        Human readable description of the operation. 0-256 characters long.
        """
        
        return super()._get_field("description", explicit_presence=False,
        )
    @description.setter
    def description(self, value: "builtins.str|None") -> None:
        return super()._set_field("description",value,explicit_presence=False,
        )
    
    @builtins.property
    def created_at(self) -> "datetime.datetime":
        """
        Creation timestamp.
        """
        
        return super()._get_field("created_at", explicit_presence=False,
        wrap=well_known_1.from_timestamp
        )
    @created_at.setter
    def created_at(self, value: "timestamp_pb2.Timestamp|datetime.datetime|None") -> None:
        return super()._set_field("created_at",value,explicit_presence=False,
        unwrap=well_known_1.to_timestamp
        )
    
    @builtins.property
    def created_by(self) -> "builtins.str":
        """
        ID of the user or service account who initiated the operation.
        """
        
        return super()._get_field("created_by", explicit_presence=False,
        )
    @created_by.setter
    def created_by(self, value: "builtins.str|None") -> None:
        return super()._set_field("created_by",value,explicit_presence=False,
        )
    
    @builtins.property
    def finished_at(self) -> "datetime.datetime":
        """
        The time when the operation finished.
        """
        
        return super()._get_field("finished_at", explicit_presence=False,
        wrap=well_known_1.from_timestamp
        )
    @finished_at.setter
    def finished_at(self, value: "timestamp_pb2.Timestamp|datetime.datetime|None") -> None:
        return super()._set_field("finished_at",value,explicit_presence=False,
        unwrap=well_known_1.to_timestamp
        )
    
    @builtins.property
    def request(self) -> "any_pb2.Any":
        """
        The request that generated this operation.
        """
        
        return super()._get_field("request", explicit_presence=False,
        )
    @request.setter
    def request(self, value: "any_pb2.Any|None") -> None:
        return super()._set_field("request",value,explicit_presence=False,
        )
    
    @builtins.property
    def request_headers(self) -> "abc.MutableMapping[builtins.str,Operation.request_header]":
        """
        The request headers that are essential for the request that generated the operation.
        For instance, ``x-resetmask``. Without these headers the request might have been processed
        differently if repeated.
        All the header names *must* be converted to lower case.
        Validator is based on:
        https://httpwg.org/specs/rfc9110.html#considerations.for.new.field.names
        """
        
        return super()._get_field("request_headers", explicit_presence=False,
        wrap=pb_classes.Map.with_wrap(Operation.request_header,None,None),
        )
    @request_headers.setter
    def request_headers(self, value: "abc.Mapping[builtins.str,Operation.request_header]|None") -> None:
        return super()._set_field("request_headers",value,explicit_presence=False,
        )
    
    @builtins.property
    def resource_id(self) -> "builtins.str":
        """
        ID of the resource that this operation creates, updates, deletes or otherwise changes.
        
        If the operation affects multiple resources or does not affect any API resources at all
        (e.g. a routine maintenance operation visible to the user), the [resource_id] must be empty.
        """
        
        return super()._get_field("resource_id", explicit_presence=False,
        )
    @resource_id.setter
    def resource_id(self, value: "builtins.str|None") -> None:
        return super()._set_field("resource_id",value,explicit_presence=False,
        )
    
    @builtins.property
    def resource(self) -> "any_pb2.Any":
        """
        .. role:: raw-html-m2r(raw)
           :format: html
        
        
        Snapshot of the resource at the moment this operation started.
        
        
        * [resource.spec] and [resource.metadata] reflect the desired resource state at the moment
          this operation started.
          E.g., in an Update operation it will be the *updated* resource spec and metadata,
          in a Create operation it will be the spec and metadata *of the resource being created*\\ ,
          and so on.
        * [resource.status] reflects the status of the resource at the moment this operation started.
          This is a snapshot, call the :raw-html-m2r:`<Resource>`\\ Service/Get to get current status of the resource.
        
        The [resource] field MUST never be updated *after* this operation has started.
        
        In a Delete operation, an operation affecting multiple resources or an operation that doesn't
        affect any API resources at all (e.g. a routine maintenance operation visible to the user),
        the [resource] inside MUST be a [google.protobuf.Empty].
        """
        
        return super()._get_field("resource", explicit_presence=False,
        )
    @resource.setter
    def resource(self, value: "any_pb2.Any|None") -> None:
        return super()._set_field("resource",value,explicit_presence=False,
        )
    
    @builtins.property
    def progress_data(self) -> "any_pb2.Any":
        """
        Additional information about the progress of an operation, e.g., a progress percentage.
        MAY be absent while the operation is running, MUST be absent after the operation has completed.
        
        Format of message inside [progress_data] is service-dependent and MUST be documented by the
        service, IF it is used.
        """
        
        return super()._get_field("progress_data", explicit_presence=False,
        )
    @progress_data.setter
    def progress_data(self, value: "any_pb2.Any|None") -> None:
        return super()._set_field("progress_data",value,explicit_presence=False,
        )
    
    @builtins.property
    def status(self) -> "request_status.RequestStatus|None":
        """
        The status of this operation. Set when this operation is completed.
        See https://github.com/grpc/grpc/blob/master/src/proto/grpc/status/status.proto.
        
        [status.code] is https://github.com/googleapis/googleapis/blob/master/google/rpc/code.proto:
        
        
        * If [status.code] == OK, the operation has completed successfully.
        * If [status.code] != OK, the operation has failed or has been cancelled.
        
          * [status.message] will contain a user-readable and actionable error message.
          * [status.details] will contain additional diagnostic information in the form of
            [ServiceError] from ../error/v1alpha1/error.proto
        
        * [status.code] must belong to an Operation-compatible subset of GRPC codes:
          OK, CANCELLED, PERMISSION_DENIED, RESOURCE_EXHAUSTED, FAILED_PRECONDITION, ABORTED, INTERNAL
        """
        
        return super()._get_field("status", explicit_presence=True,
        wrap=request_status.request_status_from_rpc_status
        )
    @status.setter
    def status(self, value: "status_pb2.Status|request_status.RequestStatus|None") -> None:
        return super()._set_field("status",value,explicit_presence=True,
        unwrap=request_status.request_status_to_rpc_status
        )
    
    __PY_TO_PB2__: builtins.dict[builtins.str,builtins.str] = {
        "id":"id",
        "description":"description",
        "created_at":"created_at",
        "created_by":"created_by",
        "finished_at":"finished_at",
        "request":"request",
        "request_headers":"request_headers",
        "resource_id":"resource_id",
        "resource":"resource",
        "progress_data":"progress_data",
        "status":"status",
        "request_header":"request_header",
        "RequestHeadersEntry":"RequestHeadersEntry",
    }
    
# file: nebius/common/v1alpha1/operation_service.proto
class GetOperationRequest(pb_classes.Message):
    __PB2_CLASS__ = operation_service_pb2.GetOperationRequest
    __PB2_DESCRIPTOR__ = descriptor.DescriptorWrap[descriptor_1.Descriptor](".nebius.common.v1alpha1.GetOperationRequest",operation_service_pb2.DESCRIPTOR,descriptor_1.Descriptor)
    __mask_functions__ = {
    }
    
    def __init__(
        self,
        initial_message: message_1.Message|None = None,
        *,
        id: "builtins.str|None|unset.UnsetType" = unset.Unset,
    ) -> None:
        super().__init__(initial_message)
        if not isinstance(id, unset.UnsetType):
            self.id = id
    
    def __dir__(self) ->abc.Iterable[builtins.str]:
        return [
            "id",
        ]
    
    @builtins.property
    def id(self) -> "builtins.str":
        """
        Operation ID.
        """
        
        return super()._get_field("id", explicit_presence=False,
        )
    @id.setter
    def id(self, value: "builtins.str|None") -> None:
        return super()._set_field("id",value,explicit_presence=False,
        )
    
    __PY_TO_PB2__: builtins.dict[builtins.str,builtins.str] = {
        "id":"id",
    }
    
class ListOperationsRequest(pb_classes.Message):
    __PB2_CLASS__ = operation_service_pb2.ListOperationsRequest
    __PB2_DESCRIPTOR__ = descriptor.DescriptorWrap[descriptor_1.Descriptor](".nebius.common.v1alpha1.ListOperationsRequest",operation_service_pb2.DESCRIPTOR,descriptor_1.Descriptor)
    __mask_functions__ = {
    }
    
    def __init__(
        self,
        initial_message: message_1.Message|None = None,
        *,
        resource_id: "builtins.str|None|unset.UnsetType" = unset.Unset,
        page_size: "builtins.int|None|unset.UnsetType" = unset.Unset,
        page_token: "builtins.str|None|unset.UnsetType" = unset.Unset,
        filter: "builtins.str|None|unset.UnsetType" = unset.Unset,
    ) -> None:
        super().__init__(initial_message)
        if not isinstance(resource_id, unset.UnsetType):
            self.resource_id = resource_id
        if not isinstance(page_size, unset.UnsetType):
            self.page_size = page_size
        if not isinstance(page_token, unset.UnsetType):
            self.page_token = page_token
        if not isinstance(filter, unset.UnsetType):
            self.filter = filter
    
    def __dir__(self) ->abc.Iterable[builtins.str]:
        return [
            "resource_id",
            "page_size",
            "page_token",
            "filter",
        ]
    
    @builtins.property
    def resource_id(self) -> "builtins.str":
        """
        ID of the Resource to list operations for.
        """
        
        return super()._get_field("resource_id", explicit_presence=False,
        )
    @resource_id.setter
    def resource_id(self, value: "builtins.str|None") -> None:
        return super()._set_field("resource_id",value,explicit_presence=False,
        )
    
    @builtins.property
    def page_size(self) -> "builtins.int":
        """
        Page size. [1...1000]. Optional, if not specified, a reasonable default will be chosen by the service.
        """
        
        return super()._get_field("page_size", explicit_presence=False,
        )
    @page_size.setter
    def page_size(self, value: "builtins.int|None") -> None:
        return super()._set_field("page_size",value,explicit_presence=False,
        )
    
    @builtins.property
    def page_token(self) -> "builtins.str":
        """
        Listing continuation token. Empty to start listing from the first page.
        """
        
        return super()._get_field("page_token", explicit_presence=False,
        )
    @page_token.setter
    def page_token(self, value: "builtins.str|None") -> None:
        return super()._set_field("page_token",value,explicit_presence=False,
        )
    
    @builtins.property
    def filter(self) -> "builtins.str":
        """
        Filter expression for the listing results. Optional.
        Filter expression format: TBD.
        """
        
        return super()._get_field("filter", explicit_presence=False,
        )
    @filter.setter
    def filter(self, value: "builtins.str|None") -> None:
        return super()._set_field("filter",value,explicit_presence=False,
        )
    
    __PY_TO_PB2__: builtins.dict[builtins.str,builtins.str] = {
        "resource_id":"resource_id",
        "page_size":"page_size",
        "page_token":"page_token",
        "filter":"filter",
    }
    
class ListOperationsResponse(pb_classes.Message):
    __PB2_CLASS__ = operation_service_pb2.ListOperationsResponse
    __PB2_DESCRIPTOR__ = descriptor.DescriptorWrap[descriptor_1.Descriptor](".nebius.common.v1alpha1.ListOperationsResponse",operation_service_pb2.DESCRIPTOR,descriptor_1.Descriptor)
    __mask_functions__ = {
    }
    
    def __init__(
        self,
        initial_message: message_1.Message|None = None,
        *,
        operations: "abc.Iterable[Operation]|None|unset.UnsetType" = unset.Unset,
        next_page_token: "builtins.str|None|unset.UnsetType" = unset.Unset,
    ) -> None:
        super().__init__(initial_message)
        if not isinstance(operations, unset.UnsetType):
            self.operations = operations
        if not isinstance(next_page_token, unset.UnsetType):
            self.next_page_token = next_page_token
    
    def __dir__(self) ->abc.Iterable[builtins.str]:
        return [
            "operations",
            "next_page_token",
        ]
    
    @builtins.property
    def operations(self) -> "abc.MutableSequence[Operation]":
        """
        List of operations on this result page.
        """
        
        return super()._get_field("operations", explicit_presence=False,
        wrap=pb_classes.Repeated.with_wrap(Operation,None,None),
        )
    @operations.setter
    def operations(self, value: "abc.Iterable[Operation]|None") -> None:
        return super()._set_field("operations",value,explicit_presence=False,
        )
    
    @builtins.property
    def next_page_token(self) -> "builtins.str":
        """
        Listing continuation token for the next page of results.
        """
        
        return super()._get_field("next_page_token", explicit_presence=False,
        )
    @next_page_token.setter
    def next_page_token(self, value: "builtins.str|None") -> None:
        return super()._set_field("next_page_token",value,explicit_presence=False,
        )
    
    __PY_TO_PB2__: builtins.dict[builtins.str,builtins.str] = {
        "operations":"operations",
        "next_page_token":"next_page_token",
    }
    
class ListOperationsByParentRequest(pb_classes.Message):
    __PB2_CLASS__ = operation_service_pb2.ListOperationsByParentRequest
    __PB2_DESCRIPTOR__ = descriptor.DescriptorWrap[descriptor_1.Descriptor](".nebius.common.v1alpha1.ListOperationsByParentRequest",operation_service_pb2.DESCRIPTOR,descriptor_1.Descriptor)
    __mask_functions__ = {
    }
    
    def __init__(
        self,
        initial_message: message_1.Message|None = None,
        *,
        parent_id: "builtins.str|None|unset.UnsetType" = unset.Unset,
        page_size: "builtins.int|None|unset.UnsetType" = unset.Unset,
        page_token: "builtins.str|None|unset.UnsetType" = unset.Unset,
        filter: "builtins.str|None|unset.UnsetType" = unset.Unset,
    ) -> None:
        super().__init__(initial_message)
        if not isinstance(parent_id, unset.UnsetType):
            self.parent_id = parent_id
        if not isinstance(page_size, unset.UnsetType):
            self.page_size = page_size
        if not isinstance(page_token, unset.UnsetType):
            self.page_token = page_token
        if not isinstance(filter, unset.UnsetType):
            self.filter = filter
    
    def __dir__(self) ->abc.Iterable[builtins.str]:
        return [
            "parent_id",
            "page_size",
            "page_token",
            "filter",
        ]
    
    @builtins.property
    def parent_id(self) -> "builtins.str":
        """
        ID of the parent to list operations for resource type at.
        """
        
        return super()._get_field("parent_id", explicit_presence=False,
        )
    @parent_id.setter
    def parent_id(self, value: "builtins.str|None") -> None:
        return super()._set_field("parent_id",value,explicit_presence=False,
        )
    
    @builtins.property
    def page_size(self) -> "builtins.int":
        """
        Page size. [1...1000]. Optional, if not specified, a reasonable default will be chosen by the service.
        """
        
        return super()._get_field("page_size", explicit_presence=False,
        )
    @page_size.setter
    def page_size(self, value: "builtins.int|None") -> None:
        return super()._set_field("page_size",value,explicit_presence=False,
        )
    
    @builtins.property
    def page_token(self) -> "builtins.str":
        """
        Listing continuation token. Empty to start listing from the first page.
        """
        
        return super()._get_field("page_token", explicit_presence=False,
        )
    @page_token.setter
    def page_token(self, value: "builtins.str|None") -> None:
        return super()._set_field("page_token",value,explicit_presence=False,
        )
    
    @builtins.property
    def filter(self) -> "builtins.str":
        """
        Filter expression for the listing results. Optional.
        Filter expression format: TBD.
        """
        
        return super()._get_field("filter", explicit_presence=False,
        )
    @filter.setter
    def filter(self, value: "builtins.str|None") -> None:
        return super()._set_field("filter",value,explicit_presence=False,
        )
    
    __PY_TO_PB2__: builtins.dict[builtins.str,builtins.str] = {
        "parent_id":"parent_id",
        "page_size":"page_size",
        "page_token":"page_token",
        "filter":"filter",
    }
    

class OperationServiceClient(client.Client):
    """
    Service for reading operations.
    
    This class provides the client methods for the ``.nebius.common.v1alpha1.OperationService`` service.
    
    Each method constructs a :class:`nebius.aio.request.Request` object
    that represents the in-flight RPC. The request can be awaited (async)
    or waited synchronously using its ``.wait()`` helpers.
    
    The request methods accept various parameters to configure metadata,
    timeouts, authorization, and retries. See individual method docstrings
    for details.
    
    :cvar __service_name__: The full protobuf service name.
    """
    
    __PB2_DESCRIPTOR__ = descriptor.DescriptorWrap[descriptor_1.ServiceDescriptor](".nebius.common.v1alpha1.OperationService",operation_service_pb2.DESCRIPTOR,descriptor_1.ServiceDescriptor)
    """The protobuf service descriptor extraction function."""
    __service_name__ = ".nebius.common.v1alpha1.OperationService"
    
    def get(self,
        request: "GetOperationRequest",
        **kwargs: typing_extensions.Unpack[request_kwargs.RequestKwargs]
    ) -> request_1.Request["GetOperationRequest","operation.Operation[Operation]"]:
        """
        Returns the latest state of the specified operation.
        
        :param request: The request object to send.
        :type request: :class:`nebius.api.nebius.common.v1alpha1.GetOperationRequest`
        
        Other parameters can be provided as keyword arguments in the
        ``**kwargs`` dictionary, including metadata, timeouts, and retries.
        See :class:`nebius.aio.request_kwargs.RequestKwargs` for details.
        
        :return: A :class:`nebius.aio.request.Request` object representing the
            in-flight RPC. It can be awaited (async) or waited
            synchronously using its ``.wait()`` helpers.
        :rtype: :class:`nebius.aio.request.Request` of
            :class:`nebius.api.nebius.common.v1alpha1.Operation`.
        """
        
        return super().request(
            method="Get",
            request=request,
            result_pb2_class=operation_pb2.Operation,
            result_wrapper=operation.Operation,
            **kwargs,
        )
    
    def list(self,
        request: "ListOperationsRequest",
        **kwargs: typing_extensions.Unpack[request_kwargs.RequestKwargs]
    ) -> request_1.Request["ListOperationsRequest","ListOperationsResponse"]:
        """
        Lists operations for the specified resource.
        
        :param request: The request object to send.
        :type request: :class:`nebius.api.nebius.common.v1alpha1.ListOperationsRequest`
        
        Other parameters can be provided as keyword arguments in the
        ``**kwargs`` dictionary, including metadata, timeouts, and retries.
        See :class:`nebius.aio.request_kwargs.RequestKwargs` for details.
        
        :return: A :class:`nebius.aio.request.Request` object representing the
            in-flight RPC. It can be awaited (async) or waited
            synchronously using its ``.wait()`` helpers.
        :rtype: :class:`nebius.aio.request.Request` of
            :class:`nebius.api.nebius.common.v1alpha1.ListOperationsResponse`.
        """
        
        return super().request(
            method="List",
            request=request,
            result_pb2_class=operation_service_pb2.ListOperationsResponse,
            result_wrapper=pb_classes.simple_wrapper(ListOperationsResponse),
            **kwargs,
        )
    

__all__ = [
    #@ local import names here @#
    "Operation",
    "GetOperationRequest",
    "ListOperationsRequest",
    "ListOperationsResponse",
    "ListOperationsByParentRequest",
    "OperationServiceClient",
]



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: nebius/compute/v1/gpu_cluster.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from nebius.api.buf.validate import validate_pb2 as buf_dot_validate_dot_validate__pb2
from nebius.api.nebius.common.v1 import metadata_pb2 as nebius_dot_common_dot_v1_dot_metadata__pb2
from nebius.api.nebius import annotations_pb2 as nebius_dot_annotations__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n#nebius/compute/v1/gpu_cluster.proto\x12\x11nebius.compute.v1\x1a\x1b\x62uf/validate/validate.proto\x1a\x1fnebius/common/v1/metadata.proto\x1a\x18nebius/annotations.proto\"\xc0\x01\n\nGpuCluster\x12>\n\x08metadata\x18\x01 \x01(\x0b\x32\".nebius.common.v1.ResourceMetadataR\x08metadata\x12\x35\n\x04spec\x18\x02 \x01(\x0b\x32!.nebius.compute.v1.GpuClusterSpecR\x04spec\x12;\n\x06status\x18\x03 \x01(\x0b\x32#.nebius.compute.v1.GpuClusterStatusR\x06status\"I\n\x0eGpuClusterSpec\x12\x37\n\x11infiniband_fabric\x18\x01 \x01(\tB\n\xbaH\x03\xc8\x01\x01\xbaJ\x01\x02R\x10infinibandFabric\"\xd5\x01\n\x10GpuClusterStatus\x12\x1c\n\tinstances\x18\x01 \x03(\tR\tinstances\x12 \n\x0breconciling\x18\x02 \x01(\x08R\x0breconciling\x12u\n\x18infiniband_topology_path\x18\x03 \x01(\x0b\x32\x39.nebius.compute.v1.GpuClusterStatusInfinibandTopologyPathH\x00R\x16infinibandTopologyPathB\n\n\x08topology\"\x89\x01\n&GpuClusterStatusInfinibandTopologyPath\x12_\n\tinstances\x18\x01 \x03(\x0b\x32\x41.nebius.compute.v1.GpuClusterStatusInfinibandTopologyPathInstanceR\tinstances\"e\n.GpuClusterStatusInfinibandTopologyPathInstance\x12\x1f\n\x0binstance_id\x18\x01 \x01(\tR\ninstanceId\x12\x12\n\x04path\x18\x02 \x03(\tR\x04pathB^\n\x18\x61i.nebius.pub.compute.v1B\x0fGpuClusterProtoP\x01Z/github.com/nebius/gosdk/proto/nebius/compute/v1b\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'nebius.compute.v1.gpu_cluster_pb2', _globals)
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\n\030ai.nebius.pub.compute.v1B\017GpuClusterProtoP\001Z/github.com/nebius/gosdk/proto/nebius/compute/v1'
  _GPUCLUSTERSPEC.fields_by_name['infiniband_fabric']._options = None
  _GPUCLUSTERSPEC.fields_by_name['infiniband_fabric']._serialized_options = b'\272H\003\310\001\001\272J\001\002'
  _globals['_GPUCLUSTER']._serialized_start=147
  _globals['_GPUCLUSTER']._serialized_end=339
  _globals['_GPUCLUSTERSPEC']._serialized_start=341
  _globals['_GPUCLUSTERSPEC']._serialized_end=414
  _globals['_GPUCLUSTERSTATUS']._serialized_start=417
  _globals['_GPUCLUSTERSTATUS']._serialized_end=630
  _globals['_GPUCLUSTERSTATUSINFINIBANDTOPOLOGYPATH']._serialized_start=633
  _globals['_GPUCLUSTERSTATUSINFINIBANDTOPOLOGYPATH']._serialized_end=770
  _globals['_GPUCLUSTERSTATUSINFINIBANDTOPOLOGYPATHINSTANCE']._serialized_start=772
  _globals['_GPUCLUSTERSTATUSINFINIBANDTOPOLOGYPATHINSTANCE']._serialized_end=873
# @@protoc_insertion_point(module_scope)



from logging import getLogger
from pathlib import Path

from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import serialization
from cryptography.hazmat.primitives.asymmetric.rsa import RSAPrivateKey
from cryptography.hazmat.primitives.asymmetric.types import PrivateKeyTypes

from .service_account import Reader as BaseReader
from .service_account import ServiceAccount

log = getLogger(__name__)


class WrongKeyTypeError(Exception):
    def __init__(self, pk: PrivateKeyTypes) -> None:
        super().__init__(f"Wrong key type {type(pk)}")


class Reader(BaseReader):
    def __init__(
        self,
        filename: str | Path,
        public_key_id: str,
        service_account_id: str,
    ) -> None:
        filename = Path(filename).expanduser()

        log.debug(f"reading SA from file {filename}")
        with open(filename, "rb") as f:
            pk = serialization.load_pem_private_key(
                f.read(), password=None, backend=default_backend()
            )
        if not isinstance(pk, RSAPrivateKey):
            raise WrongKeyTypeError(pk)
        self._pk = pk
        self._kid = public_key_id
        self._said = service_account_id

    def read(self) -> ServiceAccount:
        return ServiceAccount(self._pk, self._kid, self._said)



# type: ignore

import pytest
from pytest_subtests import SubTests

from nebius.aio import request

request.DEFAULT_AUTH_TIMEOUT = 5.0


def test_fieldkey() -> None:
    from nebius.base.fieldmask import FieldKey

    assert FieldKey("Simple_Key_123").marshal() == "Simple_Key_123"
    assert isinstance(FieldKey("abc"), FieldKey)
    assert (
        FieldKey('Key		with spaces, commas, and: ".\\_-!@$%^"').marshal()
        == '"Key\\t\\twith spaces, commas, and: \\".\\\\_-!@$%^\\""'
    )


def test_fieldpath_join(subtests) -> None:
    from nebius.base.fieldmask import FieldKey, FieldPath

    cases = [
        {
            "A": FieldPath([]),
            "B": FieldPath([]),
            "R": FieldPath([]),
        },
        {
            "A": FieldPath([FieldKey("foo")]),
            "B": FieldPath([]),
            "R": FieldPath([FieldKey("foo")]),
        },
        {
            "A": FieldPath([FieldKey("foo"), FieldKey("bar")]),
            "B": FieldPath([]),
            "R": FieldPath([FieldKey("foo"), FieldKey("bar")]),
        },
        {
            "A": FieldPath([FieldKey("foo"), FieldKey("bar")]),
            "B": FieldPath([FieldKey("baz")]),
            "R": FieldPath([FieldKey("foo"), FieldKey("bar"), FieldKey("baz")]),
        },
        {
            "A": FieldPath([FieldKey("foo")]),
            "B": FieldPath([FieldKey("bar"), FieldKey("baz")]),
            "R": FieldPath([FieldKey("foo"), FieldKey("bar"), FieldKey("baz")]),
        },
        {
            "A": FieldPath([]),
            "B": FieldPath([FieldKey("bar"), FieldKey("baz")]),
            "R": FieldPath([FieldKey("bar"), FieldKey("baz")]),
        },
    ]

    for i, case in enumerate(cases):
        with subtests.test(msg=f"case_{i}"):
            # Perform the join operation
            result = case["A"] + case["B"]
            # Assert the result matches the expected output
            assert result == case["R"], f"Failed on case {i}: {result} != {case['R']}"


def test_fieldpath_parent(subtests) -> None:
    from nebius.base.fieldmask import FieldKey, FieldPath

    # Define test cases
    cases = [
        {
            "A": FieldPath([]),
            "R": None,
        },
        {
            "A": FieldPath([FieldKey("foo")]),
            "R": FieldPath([]),
        },
        {
            "A": FieldPath([FieldKey("foo"), FieldKey("bar")]),
            "R": FieldPath([FieldKey("foo")]),
        },
        {
            "A": FieldPath([FieldKey("foo"), FieldKey("bar"), FieldKey("baz")]),
            "R": FieldPath([FieldKey("foo"), FieldKey("bar")]),
        },
    ]

    for i, case in enumerate(cases):
        with subtests.test(msg=f"case_{i}"):
            # Perform the parent operation
            result = case["A"].parent()
            # Assert the result matches the expected output
            assert result == case["R"], f"Failed on case {i}: {result} != {case['R']}"


def test_fieldpath_copy(subtests) -> None:
    from nebius.base.fieldmask import FieldKey, FieldPath

    # Define test cases
    cases = [
        {
            "A": FieldPath([]),
        },
        {
            "A": FieldPath([FieldKey("foo")]),
        },
        {
            "A": FieldPath([FieldKey("foo"), FieldKey("bar")]),
        },
        {
            "A": FieldPath([FieldKey("foo"), FieldKey("bar"), FieldKey("baz")]),
        },
    ]

    for i, case in enumerate(cases):
        with subtests.test(msg=f"case_{i}"):
            # Perform the copy operation
            cp = case["A"].copy()
            # Assert the copy matches the original
            assert cp == case["A"], f"Failed on case {i}: {cp} != {case['A']}"
            assert case["A"] == FieldPath(
                case["A"]
            ), f"Failed on case {i}: FieldPath doesn't match"  # noqa: E501
            assert case["A"] == cp, f"Failed on case {i}: A != cp"
            assert cp == case["A"], f"Failed on case {i}: cp != A"

            if len(case["A"]) > 0:
                # Modify the original and assert copy is unaffected
                case["A"][0] = FieldKey("changed")
                assert (
                    cp != case["A"]
                ), f"Failed on case {i}: cp == A after modification"
                assert (
                    case["A"] != cp
                ), f"Failed on case {i}: A equal to cp after modification"  # noqa: E501
                assert (
                    cp != case["A"]
                ), f"Failed on case {i}: cp equal to A after modification"  # noqa: E501


def test_fieldpath_equality(subtests) -> None:
    from nebius.base.fieldmask import FieldKey, FieldPath

    # Define test cases
    paths = [
        FieldPath([]),
        FieldPath([FieldKey("foo")]),
        FieldPath([FieldKey("bar")]),
        FieldPath([FieldKey("foo"), FieldKey("foo")]),
        FieldPath([FieldKey("foo"), FieldKey("bar")]),
        FieldPath([FieldKey("foo"), FieldKey("baz")]),
        FieldPath([FieldKey("foe"), FieldKey("bar")]),
        FieldPath([FieldKey("foe"), FieldKey("baz")]),
        FieldPath([FieldKey("foo"), FieldKey("bar"), FieldKey("baz")]),
        FieldPath([FieldKey("foo"), FieldKey("bae"), FieldKey("baz")]),
        FieldPath([FieldKey("foe"), FieldKey("bar"), FieldKey("baz")]),
        FieldPath([FieldKey("foo"), FieldKey("bar"), FieldKey("bax")]),
        FieldPath([FieldKey("foo"), FieldKey("foo"), FieldKey("bar")]),
        FieldPath([FieldKey("foo"), FieldKey("foo"), FieldKey("baz")]),
        FieldPath([FieldKey("foo"), FieldKey("foo"), FieldKey("foo")]),
    ]

    for i, a in enumerate(paths):
        for j, b in enumerate(paths):
            with subtests.test(msg=f"case_{i}_{j}"):
                # Test equality
                assert a == a, f"Self-equality failed for case {i}"
                assert b == b, f"Self-equality failed for case {j}"

                if i == j:
                    assert a == b, f"Equality failed for cases {i}, {j}"
                    assert b == a, f"Equality symmetry failed for cases {i}, {j}"
                else:
                    assert a != b, f"Inequality failed for cases {i}, {j}"
                    assert b != a, f"Inequality symmetry failed for cases {i}, {j}"


def test_fieldpath_to_mask(subtests):
    from nebius.base.fieldmask import FieldKey, FieldPath, Mask

    # Define test cases
    cases = [
        {"FP": FieldPath([]), "Res": ""},
        {"FP": FieldPath([FieldKey("foo")]), "Res": "foo"},
        {"FP": FieldPath([FieldKey("foo"), FieldKey("bar")]), "Res": "foo.bar"},
        {
            "FP": FieldPath([FieldKey("foo"), FieldKey("bar"), FieldKey("baz")]),
            "Res": "foo.bar.baz",
        },
    ]

    for i, case in enumerate(cases):
        with subtests.test(msg=f"case_{i}"):
            # Convert FieldPath to Mask
            m = case["FP"].to_mask()

            # Check that the result is an instance of Mask
            assert isinstance(
                m, Mask
            ), f"Failed on case {i}: Result is not a Mask instance"

            # Marshal the mask to a string
            str_representation = m.marshal()

            # Assert no error and correct string representation
            assert str_representation == case["Res"], (
                f"Failed on case {i}: " f"{str_representation} !=" f" {case['Res']}"
            )


def test_fieldpath_is_prefix_of(subtests):
    from nebius.base.fieldmask import FieldKey, FieldPath

    # Define test cases
    cases = [
        {"A": FieldPath([]), "B": FieldPath([]), "Res": False},
        {"A": FieldPath([FieldKey("foo")]), "B": FieldPath([]), "Res": False},
        {
            "A": FieldPath([FieldKey("foo")]),
            "B": FieldPath([FieldKey("bar")]),
            "Res": False,
        },  # noqa: E501
        {
            "A": FieldPath([FieldKey("foo"), FieldKey("baz")]),
            "B": FieldPath([FieldKey("foo"), FieldKey("bar")]),
            "Res": False,
        },  # noqa: E501
        {
            "A": FieldPath([FieldKey("foo"), FieldKey("baz"), FieldKey("abc")]),
            "B": FieldPath(
                [FieldKey("foo"), FieldKey("bar"), FieldKey("abc"), FieldKey("def")]
            ),
            "Res": False,
        },  # noqa: E501
        {
            "A": FieldPath([FieldKey("baz"), FieldKey("foo")]),
            "B": FieldPath(
                [FieldKey("bar"), FieldKey("foo"), FieldKey("abc"), FieldKey("def")]
            ),
            "Res": False,
        },  # noqa: E501
        {
            "A": FieldPath([FieldKey("baz"), FieldKey("foo"), FieldKey("abc")]),
            "B": FieldPath(
                [FieldKey("bar"), FieldKey("foo"), FieldKey("abc"), FieldKey("def")]
            ),
            "Res": False,
        },  # noqa: E501
        {
            "A": FieldPath([FieldKey("bar"), FieldKey("foo"), FieldKey("")]),
            "B": FieldPath(
                [FieldKey("bar"), FieldKey("foo"), FieldKey("abc"), FieldKey("def")]
            ),
            "Res": False,
        },  # noqa: E501
        {"A": FieldPath([]), "B": FieldPath([FieldKey("foo")]), "Res": True},
        {
            "A": FieldPath([FieldKey("bar")]),
            "B": FieldPath([FieldKey("bar"), FieldKey("foo")]),
            "Res": True,
        },  # noqa: E501
        {
            "A": FieldPath([FieldKey("bar"), FieldKey("baz")]),
            "B": FieldPath([FieldKey("bar"), FieldKey("baz"), FieldKey("foo")]),
            "Res": True,
        },  # noqa: E501
        {
            "A": FieldPath([]),
            "B": FieldPath([FieldKey("foo"), FieldKey("bar"), FieldKey("baz")]),
            "Res": True,
        },  # noqa: E501
        {
            "A": FieldPath([FieldKey("bar")]),
            "B": FieldPath([FieldKey("bar"), FieldKey("foo"), FieldKey("baz")]),
            "Res": True,
        },  # noqa: E501
        {
            "A": FieldPath([FieldKey("bar"), FieldKey("baz")]),
            "B": FieldPath(
                [FieldKey("bar"), FieldKey("baz"), FieldKey("foo"), FieldKey("abc")]
            ),
            "Res": True,
        },  # noqa: E501
        {
            "A": FieldPath([FieldKey("bar")]),
            "B": FieldPath(
                [FieldKey("bar"), FieldKey("baz"), FieldKey("foo"), FieldKey("abc")]
            ),
            "Res": True,
        },  # noqa: E501
        {
            "A": FieldPath([]),
            "B": FieldPath(
                [FieldKey("bar"), FieldKey("baz"), FieldKey("foo"), FieldKey("abc")]
            ),
            "Res": True,
        },  # noqa: E501
    ]

    for i, case in enumerate(cases):
        with subtests.test(msg=f"case_{i}"):
            res = case["A"].is_prefix_of(case["B"])
            assert res == case["Res"], f"Failed on case {i}: {res} != {case['Res']}"


def test_fieldpath_matches_reset_mask(subtests) -> None:
    from nebius.base.fieldmask import FieldKey, FieldPath, Mask

    cases = [
        {
            "FP": FieldPath([]),
            "M": None,
            "Res": False,
            "Final": False,
        },
        {
            "FP": FieldPath([FieldKey("abc")]),
            "M": None,
            "Res": False,
            "Final": False,
        },
        {
            "FP": FieldPath([]),
            "M": Mask(),
            "Res": True,
            "Final": True,
        },
        {
            "FP": FieldPath([FieldKey("abc")]),
            "M": Mask(),
            "Res": False,
            "Final": False,
        },
        {
            "FP": FieldPath([FieldKey("abc")]),
            "M": Mask(field_parts={FieldKey("abc"): Mask()}),
            "Res": True,
            "Final": True,
        },
        {
            "FP": FieldPath([FieldKey("abc")]),
            "M": Mask(field_parts={FieldKey("abc"): Mask(any=Mask())}),
            "Res": True,
            "Final": False,
        },
        {
            "FP": FieldPath([FieldKey("abc")]),
            "M": Mask(
                field_parts={
                    FieldKey("abc"): Mask(field_parts={FieldKey("foo"): Mask()})
                }
            ),
            "Res": True,
            "Final": False,
        },
        {
            "FP": FieldPath([FieldKey("abc")]),
            "M": Mask(
                field_parts={
                    FieldKey("abc"): Mask(field_parts={FieldKey("foo"): Mask()}),
                    FieldKey("def"): Mask(field_parts={FieldKey("foo"): Mask()}),
                }
            ),
            "Res": True,
            "Final": False,
        },
        {
            "FP": FieldPath([FieldKey("abc"), FieldKey("foo")]),
            "M": Mask(
                any=Mask(field_parts={FieldKey("foo"): Mask()}),
                field_parts={
                    FieldKey("abc"): Mask(field_parts={FieldKey("bar"): Mask()}),
                    FieldKey("def"): Mask(field_parts={FieldKey("baz"): Mask()}),
                },
            ),
            "Res": True,
            "Final": True,
        },
        {
            "FP": FieldPath([FieldKey("abc"), FieldKey("bar")]),
            "M": Mask(
                any=Mask(field_parts={FieldKey("bar"): Mask(any=Mask())}),
                field_parts={
                    FieldKey("abc"): Mask(field_parts={FieldKey("bar"): Mask()}),
                    FieldKey("def"): Mask(field_parts={FieldKey("baz"): Mask()}),
                },
            ),
            "Res": True,
            "Final": True,
        },
        {
            "FP": FieldPath([FieldKey("abc"), FieldKey("bar")]),
            "M": Mask(
                any=Mask(field_parts={FieldKey("bar"): Mask()}),
                field_parts={
                    FieldKey("abc"): Mask(
                        field_parts={FieldKey("bar"): Mask(any=Mask())}
                    ),
                    FieldKey("def"): Mask(field_parts={FieldKey("baz"): Mask()}),
                },
            ),
            "Res": True,
            "Final": True,
        },
        {
            "FP": FieldPath([FieldKey("abc"), FieldKey("bar")]),
            "M": Mask(
                any=Mask(field_parts={FieldKey("bar"): Mask()}),
                field_parts={
                    FieldKey("abc"): Mask(field_parts={FieldKey("bar"): Mask()}),
                    FieldKey("def"): Mask(field_parts={FieldKey("baz"): Mask()}),
                },
            ),
            "Res": True,
            "Final": True,
        },
        {
            "FP": FieldPath([FieldKey("abc")]),
            "M": Mask(any=Mask()),
            "Res": True,
            "Final": True,
        },
        {
            "FP": FieldPath([FieldKey("abc")]),
            "M": Mask(any=Mask(any=Mask())),
            "Res": True,
            "Final": False,
        },
        {
            "FP": FieldPath([FieldKey("x")]),
            "M": Mask(
                any=Mask(),
                field_parts={FieldKey("x"): Mask(field_parts={FieldKey("y"): Mask()})},
            ),
            "Res": True,
            "Final": True,
        },
        {
            "FP": FieldPath([FieldKey("x")]),
            "M": Mask(
                any=Mask(field_parts={FieldKey("y"): Mask()}),
                field_parts={FieldKey("x"): Mask()},
            ),
            "Res": True,
            "Final": True,
        },
    ]

    for i, case in enumerate(cases):
        with subtests.test(msg=f"case_{i}"):
            # Test the MatchesResetMask method
            res = case["FP"].matches_reset_mask(case["M"])
            res_final = case["FP"].matches_reset_mask_final(case["M"])
            assert res == case["Res"], f"Failed on case {i}: {res} != {case['Res']}"
            assert (
                res_final == case["Final"]
            ), f"Failed on case {i}: {res} != {case['Final']}"  # noqa: E501


def test_fieldpath_matches_select_mask(subtests) -> None:
    from nebius.base.fieldmask import FieldKey, FieldPath, Mask

    cases = [
        {
            "FP": FieldPath([]),
            "M": None,
            "Res": True,
        },
        {
            "FP": FieldPath([FieldKey("foo")]),
            "M": None,
            "Res": True,
            "Inner": True,
        },
        {
            "FP": FieldPath([FieldKey("foo"), FieldKey("bar")]),
            "M": None,
            "Res": True,
            "Inner": True,
        },
        {
            "FP": FieldPath([FieldKey("foo"), FieldKey("bar")]),
            "M": Mask(),
            "Res": True,
            "Inner": True,
        },
        {
            "FP": FieldPath([FieldKey("foo")]),
            "M": Mask(),
            "Res": True,
            "Inner": True,
        },
        {
            "FP": FieldPath([]),
            "M": Mask(),
            "Res": True,
        },
        {
            "FP": FieldPath([]),
            "M": Mask.unmarshal("a,b.c"),
            "Res": True,
        },
        {
            "FP": FieldPath([FieldKey("a")]),
            "M": Mask.unmarshal("a,b.c"),
            "Res": True,
        },
        {
            "FP": FieldPath([FieldKey("b")]),
            "M": Mask.unmarshal("a,b.c"),
            "Res": True,
        },
        {
            "FP": FieldPath([FieldKey("b"), FieldKey("c")]),
            "M": Mask.unmarshal("a,b.c"),
            "Res": True,
        },
        {
            "FP": FieldPath([FieldKey("b"), FieldKey("c"), FieldKey("d")]),
            "M": Mask.unmarshal("a,b.c"),
            "Res": True,
            "Inner": True,
        },
        {
            "FP": FieldPath([FieldKey("a"), FieldKey("c"), FieldKey("d")]),
            "M": Mask.unmarshal("a,b.c"),
            "Res": True,
            "Inner": True,
        },
        {
            "FP": FieldPath([FieldKey("b"), FieldKey("d")]),
            "M": Mask.unmarshal("a,b.c"),
            "Res": False,
        },
        {
            "FP": FieldPath([FieldKey("d")]),
            "M": Mask.unmarshal("a,b.c"),
            "Res": False,
        },
        {
            "FP": FieldPath([FieldKey("a"), FieldKey("d")]),
            "M": Mask.unmarshal("a.*"),
            "Res": True,
        },
        {
            "FP": FieldPath([FieldKey("a"), FieldKey("d"), FieldKey("e")]),
            "M": Mask.unmarshal("a.*"),
            "Res": True,
            "Inner": True,
        },
        {
            "FP": FieldPath([FieldKey("a"), FieldKey("d"), FieldKey("e")]),
            "M": Mask.unmarshal("a.*.c"),
            "Res": False,
        },
        {
            "FP": FieldPath([FieldKey("a"), FieldKey("d"), FieldKey("c")]),
            "M": Mask.unmarshal("a.*.c"),
            "Res": True,
        },
    ]

    for i, case in enumerate(cases):
        with subtests.test(msg=f"case_{i}"):
            # Test the MatchesSelectMask method
            res = case["FP"].matches_select_mask(case["M"])
            res2, inner = case["FP"].matches_select_mask_inner(case["M"])

            # Assert the results
            assert res == case["Res"], f"Failed on case {i}: {res} != {case['Res']}"
            assert res2 == case["Res"], f"Failed on case {i}: {res2} != {case['Res']}"
            assert inner == case.get(
                "Inner", False
            ), f"Failed on case {i}: {inner} != {case.get('Inner', False)}"  # noqa: E501


def test_fieldpath_marshal(subtests) -> None:
    from nebius.base.fieldmask import FieldKey, FieldPath

    cases = [
        {
            "FP": FieldPath([]),
            "M": "",
        },
        {
            "FP": FieldPath([FieldKey("foo")]),
            "M": "foo",
        },
        {
            "FP": FieldPath([FieldKey("foo"), FieldKey("bar")]),
            "M": "foo.bar",
        },
        {
            "FP": FieldPath([FieldKey("foo.bar")]),
            "M": '"foo.bar"',
        },
        {
            "FP": FieldPath([FieldKey("foo.bar"), FieldKey("baz")]),
            "M": '"foo.bar".baz',
        },
    ]

    for i, case in enumerate(cases):
        with subtests.test(msg=f"case_{i}"):
            res = case["FP"].marshal()
            assert (
                res == case["M"]
            ), f"Failed on case {i}: expected {case['M']}, got {res}"


def test_fieldpath_unmarshal(subtests) -> None:
    from nebius.base.fieldmask import Error, FieldKey, FieldPath

    cases = [
        {
            "NoStarter": True,
            "Err": "wrong type of source\\: \\<class \\'NoneType\\'\\>, expected str",
        },
        {
            "Mask": "(",
            "Err": "unclosed left brace at position 0 near",
        },
        {
            "Mask": "a,b",
            "Err": "multiple paths in the mask",
        },
        {
            "Mask": "a.b",
            "Result": FieldPath([FieldKey("a"), FieldKey("b")]),
        },
    ]

    for i, case in enumerate(cases):
        with subtests.test(msg=f"case_{i}"):
            mask = None if case.get("NoStarter", False) else case["Mask"]
            if "Err" in case:
                with pytest.raises((Error), match=case["Err"]):
                    FieldPath.unmarshal(mask)
            else:
                res = FieldPath.unmarshal(mask)
                assert (
                    res == case["Result"]
                ), f"Failed on case {i}: expected {case['Result']}, got {res}"  # noqa: E501


def test_parse_fieldmask(subtests) -> None:
    from nebius.base.fieldmask import Mask
    from nebius.base.fieldmask_parser import ParseError, parse

    cases = [
        {"Input": "*,", "Err": r"unexpected end of mask"},
        {"Input": "*.", "Err": r"unexpected end of mask"},
        {
            "Input": "*.(",
            "Err": r"unclosed left brace at position 2 near \"\*\.\u20de\(\"",
        },  # noqa: E501
        {
            "Input": " \t\r\n*.( \t\r\n",
            "Err": r"unclosed left brace at position 6 near \" \\t\\r\\n\*\.\u20de\( \\t\\r\\n\"",  # noqa: E501
        },
        {
            "Input": "abcdefghijklmonpqrst.(abcdefghijklmonpqrst",
            "Err": r"unclosed left brace at position 21 near \"jklmonpqrst\.\u20de\(abcdefghijklmo\.\.\.\"",  # noqa: E501
        },
        {
            "Input": "abcdefghijklmonpqrst)abcdefghijklmonpqrst",
            "Err": r"unmatched right brace at position 20 near \"ijklmonpqrst\u20de\)abcdefghijklmo\.\.\.\"",  # noqa: E501
        },
        {
            "Input": "abcdefghijklmonpqrst..abcdefghijklmonpqrst",
            "Err": r"unexpected token TokenDOT\(\"\.\" pos 21\), expecting field or submask at position 21 near \"jklmonpqrst\.\u20de\.abcdefghijklmo\.\.\.\"",  # noqa: E501
        },
        {
            "Input": "abcdefghijklmonpqrst abcdefghijklmonpqrst feafwafwawfadwadw",
            "Err": r"unexpected token TokenPLAIN_KEY\(\"abcdefghijklmonpqrst\" pos 21\), expecting separator or closing brace at position 21 near \"jklmonpqrst \u20deabcdefghijklmon\.\.\.\"",  # noqa: E501
        },
        {"Input": "#", "Err": r"unexpected symbol at position 0 near \"\u20de#\""},
        {
            "Input": "#1234567890abcdefghijklmnopqrst",
            "Err": r"unexpected symbol at position 0 near \"\u20de#1234567890abcdefghijklmnop\.\.\.\"",  # noqa: E501
        },
        {
            "Input": '"1234567890abcdefghijklmnopqrst',
            "Err": r"unterminated quoted string at position 0 near \"\u20de\\\"1234567890abcdefghijklmnop\.\.\.\"",  # noqa: E501
        },
        {"Input": "", "Output": ""},
        {"Input": "()", "Output": ""},
        {"Input": " \t\r\n( \t\r\n) \t\r\n", "Output": ""},
        {"Input": " \t\r\n", "Output": ""},
        {"Input": "*", "Output": "*"},
        {"Input": " \t\r\n* \t\r\n", "Output": "*"},
        {"Input": " \t\r\na \t\r\n", "Output": "a"},
        {"Input": " \t\r\n( \t\r\na \t\r\n) \t\r\n", "Output": "a"},
        {"Input": "*,*,*,*", "Output": "*"},
        {"Input": "test", "Output": "test"},
        {"Input": '"test"', "Output": "test"},
        {"Input": '"test "', "Output": '"test "'},
        {"Input": "a,a", "Output": "a"},
        {"Input": "a.b", "Output": "a.b"},
        {"Input": "a \t\r\n. \t\r\nb", "Output": "a.b"},
        {"Input": "a \t\r\n, \t\r\na", "Output": "a"},
        {"Input": "*,test", "Output": "*,test"},
        {"Input": "* \t\r\n, \t\r\ntest", "Output": "*,test"},
        {"Input": "test,*", "Output": "*,test"},
        {"Input": "a.b,a.b", "Output": "a.b"},
        {"Input": "a.*,a.*", "Output": "a.*"},
        {"Input": "*.b,*.b", "Output": "*.b"},
        {"Input": "a.b,a.c", "Output": "a.(b,c)"},
        {"Input": "a.(b)", "Output": "a.b"},
        {"Input": "a.(b,())", "Output": "a.b"},
        {"Input": "a.((),b)", "Output": "a.b"},
        {"Input": "a.((()))", "Output": "a"},
        {"Input": "a.(b,c)", "Output": "a.(b,c)"},
        {"Input": "*.(b,c)", "Output": "*.(b,c)"},
        {"Input": "a.(b,c).(d,e)", "Output": "a.(b.(d,e),c.(d,e))"},
        {"Input": "a.(*,c).(d,e)", "Output": "a.(*.(d,e),c.(d,e))"},
        {"Input": "a.(((((((*,c))))))).(d,e)", "Output": "a.(*.(d,e),c.(d,e))"},
        {
            "Input": "(((((((a))))))).(((((*)))),(((c)))).(((((((d,(((e))))))))))",
            "Output": "a.(*.(d,e),c.(d,e))",
        },
        {"Input": "a.(b,c.(d,e))", "Output": "a.(b,c.(d,e))"},
        {"Input": "a.(*,b,c)", "Output": "a.(*,b,c)"},
        {"Input": "a.(*,b,c.(d,e))", "Output": "a.(*,b,c.(d,e))"},
        {
            "Input": "*.*.(a,b,c.*,d.(e,f),g.(*.(h,i),j,k)),1,A,B,l,m,n.*,o.(p,q,w,x),r.(*.(s,t),u.*,v),z.(*.y,u.*,v)",  # noqa: E501
            "Output": "*.*.(a,b,c.*,d.(e,f),g.(*.(h,i),j,k)),1,A,B,l,m,n.*,o.(p,q,w,x),r.(*.(s,t),u.*,v),z.(*.y,u.*,v)",  # noqa: E501
        },
        {
            "Input": 'a."\\",.() \\\\t\\\\r\\\\n".b',
            "Output": 'a."\\",.() \\\\t\\\\r\\\\n".b',
        },  # noqa: E501
    ]

    for i, case in enumerate(cases):
        with subtests.test(msg=f"case_{i}"):
            if "Err" in case:
                with pytest.raises(ParseError, match=case["Err"]):
                    parse(case["Input"])
            else:
                result = parse(case["Input"])
                assert isinstance(result, Mask)
                normalized = result.marshal()
                assert (
                    normalized == case["Output"]
                ), f"Failed on case {i}: expected {case['Output']}, got {normalized}"  # noqa: E501


def test_mask_is_empty() -> None:
    from nebius.base.fieldmask import FieldKey, Mask

    assert Mask().is_empty()
    assert Mask(any=None).is_empty()
    assert Mask(field_parts=dict[FieldKey, Mask]()).is_empty()
    assert Mask(any=None, field_parts=dict[FieldKey, Mask]()).is_empty()
    assert Mask(any=None, field_parts=dict[FieldKey, Mask]()).is_empty()
    assert not Mask(any=Mask()).is_empty()
    assert not Mask(field_parts=dict[FieldKey, Mask]({FieldKey(""): Mask()})).is_empty()


def test_mask_marshal(subtests: SubTests) -> None:
    from nebius.base.fieldmask import Mask

    infinite_mask = Mask()
    infinite_mask.any = infinite_mask

    cases = [
        {
            "Mask": Mask(),
            "Result": "",
        },
        {
            "Mask": Mask(),
            "Result": "",
        },
        {
            "Mask": infinite_mask,
            "Err": RecursionError,
        },
        {
            "Mask": Mask(
                field_parts={
                    "test": infinite_mask,
                },
            ),
            "Err": RecursionError,
        },
        {
            "Mask": Mask(
                any=Mask(),
            ),
            "Result": "*",
        },
        {
            "Mask": Mask(any=Mask()),
            "Result": "*",
        },
        {
            "Mask": Mask(
                field_parts={
                    "test": Mask(),
                }
            ),
            "Result": "test",
        },
        {
            "Mask": Mask(
                field_parts=Mask(),
            ),
            "Result": "",
        },
        {
            "Mask": Mask(
                any=Mask(),
                field_parts={
                    "test": Mask(),
                },
            ),
            "Result": "*,test",
        },
        {
            "Mask": Mask(
                field_parts={
                    "test": Mask(),
                    "foo": Mask(),
                },
            ),
            "Result": "foo,test",
        },
        {
            "Mask": Mask(
                any=Mask(),
                field_parts={
                    "test": Mask(),
                    "foo": Mask(),
                },
            ),
            "Result": "*,foo,test",
        },
        {
            "Mask": Mask(
                any=Mask(
                    any=Mask(),
                ),
            ),
            "Result": "*.*",
        },
        {
            "Mask": Mask(
                any=Mask(
                    any=Mask(),
                    field_parts={
                        "test": Mask(),
                    },
                ),
            ),
            "Result": "*.(*,test)",
        },
        {
            "Mask": Mask(
                any=Mask(
                    any=Mask(),
                    field_parts={
                        "test": Mask(
                            any=Mask(),
                        ),
                    },
                ),
            ),
            "Result": "*.(*,test.*)",
        },
        {
            "Mask": Mask(
                any=Mask(
                    field_parts={
                        "test": Mask(
                            any=Mask(),
                        ),
                    },
                ),
            ),
            "Result": "*.test.*",
        },
        {
            "Mask": Mask(
                any=Mask(
                    any=Mask(
                        any=Mask(),
                    ),
                ),
            ),
            "Result": "*.*.*",
        },
        {
            "Mask": Mask(
                field_parts={
                    "test": Mask(
                        any=Mask(),
                    ),
                },
            ),
            "Result": "test.*",
        },
        {
            "Mask": Mask(
                any=Mask(
                    field_parts={
                        "test": Mask(),
                    },
                ),
            ),
            "Result": "*.test",
        },
        {
            "Mask": Mask(
                field_parts={
                    "test": Mask(
                        any=Mask(),
                    ),
                },
            ),
            "Result": "test.*",
        },
        {
            "Mask": Mask(
                field_parts={
                    "test": Mask(
                        field_parts={
                            "inner": Mask(),
                        },
                    ),
                },
            ),
            "Result": "test.inner",
        },
        {
            "Mask": Mask(
                field_parts={
                    "test.inner": Mask(
                        field_parts={
                            "inner": Mask(),
                        },
                    ),
                },
            ),
            "Result": '"test.inner".inner',
        },
        {
            "Mask": Mask(
                field_parts={
                    "test,inner": Mask(
                        field_parts={
                            "inner": Mask(),
                        },
                    ),
                },
            ),
            "Result": '"test,inner".inner',
        },
        {
            "Mask": Mask(
                field_parts={
                    "test(inner)": Mask(
                        field_parts={
                            "inner": Mask(),
                        },
                    ),
                },
            ),
            "Result": '"test(inner)".inner',
        },
        {
            "Mask": Mask(
                field_parts={
                    '"test"': Mask(),
                },
            ),
            "Result": '"\\"test\\""',
        },
        {
            "Mask": Mask(
                any=Mask(),
                field_parts={
                    '"test"': Mask(),
                },
            ),
            "Result": '"\\"test\\"",*',
        },
        {
            "Mask": Mask(
                any=Mask(),
                field_parts={
                    '",.() ' + "\t\r\n": Mask(),
                },
            ),
            "Result": '"\\",.() \\t\\r\\n",*',
        },
        {
            "Mask": Mask(
                any=Mask(),
                field_parts={
                    '"test"': Mask(
                        field_parts={
                            "inner": Mask(),
                        },
                    ),
                },
            ),
            "Result": '"\\"test\\"".inner,*',
        },
        {
            "Mask": Mask(
                any=Mask(),
                field_parts={
                    '"test"': Mask(
                        field_parts={
                            "inner": Mask(),
                            "*": Mask(),
                        },
                    ),
                },
            ),
            "Result": '"\\"test\\"".("*",inner),*',
        },
        {
            "Mask": Mask(
                any=Mask(
                    any=Mask(
                        field_parts={
                            "a": Mask(),
                            "b": Mask(),
                            "c": Mask(
                                any=Mask(),
                            ),
                            "d": Mask(
                                field_parts={
                                    "e": Mask(),
                                    "f": Mask(),
                                },
                            ),
                            "g": Mask(
                                any=Mask(
                                    field_parts={
                                        "h": Mask(),
                                        "i": Mask(),
                                    },
                                ),
                                field_parts={
                                    "j": Mask(),
                                    "k": Mask(),
                                },
                            ),
                        },
                    ),
                ),
                field_parts={
                    "l": Mask(),
                    "m": Mask(),
                    "n": Mask(
                        any=Mask(),
                    ),
                    "o": Mask(
                        field_parts={
                            "p": Mask(),
                            "q": Mask(),
                        },
                    ),
                    "r": Mask(
                        any=Mask(
                            field_parts={
                                "s": Mask(),
                                "t": Mask(),
                            },
                        ),
                        field_parts={
                            "u": Mask(
                                any=Mask(),
                            ),
                            "v": Mask(),
                        },
                    ),
                },
            ),
            "Result": "*.*.(a,b,c.*,d.(e,f),g.(*.(h,i),j,k)),l,m,n.*,o.(p,q),r.(*.(s,t),u.*,v)",  # noqa: E501
        },
        {
            "Mask": Mask(
                any=Mask(
                    any=Mask(
                        field_parts={
                            "a": Mask(),
                            '"b"': Mask(),
                            '"c"': Mask(
                                any=Mask(),
                            ),
                            '"d"': Mask(
                                field_parts={
                                    "e": Mask(),
                                    '"f"': Mask(),
                                },
                            ),
                            "g": Mask(
                                any=Mask(
                                    field_parts={
                                        '"h"': Mask(),
                                        "i": Mask(),
                                    },
                                ),
                                field_parts={
                                    '"j"': Mask(),
                                    "k": Mask(),
                                },
                            ),
                        },
                    ),
                ),
                field_parts={
                    "l": Mask(),
                    "m": Mask(),
                    "n": Mask(
                        any=Mask(),
                    ),
                    '"o"': Mask(
                        field_parts={
                            '"p"': Mask(),
                            "q": Mask(),
                            "*": Mask(),
                        },
                    ),
                    "r": Mask(
                        any=Mask(
                            field_parts={
                                "s": Mask(),
                                "t": Mask(),
                                "*": Mask(),
                            },
                        ),
                        field_parts={
                            "u": Mask(
                                any=Mask(),
                            ),
                            "v": Mask(),
                            "*": Mask(),
                        },
                    ),
                },
            ),
            "Result": '"\\"o\\"".("*","\\"p\\"",q),*.*.("\\"b\\"","\\"c\\"".*,"\\"d\\"".("\\"f\\"",e),a,g.("\\"j\\"",*.("\\"h\\"",i),k)),l,m,n.*,r.("*",*.("*",s,t),u.*,v)',  # noqa: E501
        },
    ]

    for i, case in enumerate(cases):
        with subtests.test(msg=f"case_{i}"):
            mask = case["Mask"]
            expected_result = case.get("Result")
            expected_err = case.get("Err")

            if expected_err:
                with pytest.raises(expected_err):
                    mask.marshal()
            else:
                result = mask.marshal()
                assert result == expected_result, f"Failed on case {i}"


def test_mask_unmarshal_text(subtests: SubTests) -> None:
    from nebius.base.fieldmask import Error, Mask

    cases = [
        {"mask": "(", "err": 'unclosed left brace at position 0 near "\u20de("'},
        {"mask": "", "result": Mask()},
        {"mask": "a", "result": Mask(field_parts={"a": Mask()})},
        {
            "mask": "test.inner",
            "result": Mask(field_parts={"test": Mask(field_parts={"inner": Mask()})}),
        },
        {
            "mask": "*",
            "result": Mask(any=Mask()),
        },
        {
            "mask": "*.test",
            "result": Mask(any=Mask(field_parts={"test": Mask()})),
        },
        {
            "mask": "test.*",
            "result": Mask(field_parts={"test": Mask(any=Mask())}),
        },
        {
            "mask": "test.(inner,outer)",
            "result": Mask(
                field_parts={
                    "test": Mask(field_parts={"inner": Mask(), "outer": Mask()})
                }
            ),
        },
        {
            "mask": "*.*",
            "result": Mask(any=Mask(any=Mask())),
        },
    ]

    for i, case in enumerate(cases):
        with subtests.test(msg=f"case_{i}"):
            expected_result = case.get("result")
            expected_err = case.get("err")

            if expected_err:
                with pytest.raises(Error) as exc_info:
                    Mask.unmarshal(case["mask"])
                assert str(exc_info.value) == str(expected_err), f"Failed on case {i}"
            else:
                mask = Mask.unmarshal(case["mask"])
                assert mask == expected_result, f"Failed on case {i}"


def test_mask_equality(subtests: SubTests) -> None:
    from nebius.base.fieldmask import Mask

    infinite_mask = Mask()
    infinite_mask.any = infinite_mask

    assert infinite_mask != infinite_mask
    assert Mask() != ""
    assert "" != Mask()

    # Test cases for equality
    masks = [
        Mask(),
        Mask(any=Mask()),
        Mask(any=Mask(any=Mask())),
        Mask(any=Mask(field_parts={"a": Mask()})),
        Mask(field_parts={"a": Mask()}),
        Mask(field_parts={"a": Mask(any=Mask())}),
        Mask(field_parts={"b": Mask()}),
        Mask(field_parts={"a": Mask(), "b": Mask()}),
        Mask(any=Mask(), field_parts={"a": Mask()}),
        Mask(any=Mask(), field_parts={"b": Mask()}),
        Mask(any=Mask(), field_parts={"a": Mask(), "b": Mask()}),
    ]

    for i, mask1 in enumerate(masks):
        with subtests.test(msg=f"mask_{i}"):
            assert mask1 == mask1
            assert mask1 == mask1.copy()
            assert mask1.copy() == mask1
            assert mask1 != infinite_mask
            assert infinite_mask != mask1
            for j, mask2 in enumerate(masks):
                with subtests.test(msg=f"mask_{i}_vs_mask_{j}"):
                    if i != j:
                        assert mask1 != mask2, f"mask{i} must not be equal to mask{j}"
                        assert mask2 != mask1, f"mask{j} must not be equal to mask{i}"
                        assert (
                            mask2.copy() != mask1
                        ), f"mask{j} copy must not be equal to mask{i}"  # noqa: E501
                        assert (
                            mask1.copy() != mask2
                        ), f"mask{i} copy must not be equal to mask{j}"  # noqa: E501
                        assert (
                            mask1 != mask2.copy()
                        ), f"mask{i} must not be equal to mask{j} copy"  # noqa: E501
                        assert (
                            mask2 != mask1.copy()
                        ), f"mask{j} must not be equal to mask{i} copy"  # noqa: E501


def test_mask_copy() -> None:
    from nebius.base.fieldmask import Mask

    infinite_mask = Mask()
    infinite_mask.any = infinite_mask

    with pytest.raises(RecursionError):
        infinite_mask.copy()
    with pytest.raises(RecursionError):
        Mask(any=infinite_mask).copy()
    with pytest.raises(RecursionError):
        Mask(field_parts={"a": infinite_mask}).copy()

    m = Mask()
    c = m.copy()
    assert c == m
    m.any = Mask()
    assert c != m

    m = Mask(any=Mask())
    c = m.copy()
    assert c == m
    m.any.any = Mask()
    assert c != m

    m = Mask(field_parts={"a": Mask()})
    c = m.copy()
    assert c == m
    m.any = Mask()
    assert c != m

    m = Mask(field_parts={"a": Mask()})
    c = m.copy()
    assert c == m
    m.field_parts["c"] = Mask()
    assert c != m

    m = Mask(field_parts={"a": Mask()})
    c = m.copy()
    assert c == m
    m.field_parts["a"] = Mask(any=Mask())
    assert c != m

    m = Mask(field_parts={"a": Mask()})
    c = m.copy()
    assert c == m
    m.field_parts["a"].any = Mask()
    assert c != m


def test_mask_merge_infinite_recursion(subtests) -> None:
    from nebius.base.fieldmask import Mask

    infinite_mask = Mask()
    infinite_mask.any = infinite_mask

    cases = [
        (infinite_mask, infinite_mask, RecursionError),
        (Mask(), infinite_mask, RecursionError),
        (Mask(), Mask(any=infinite_mask), RecursionError),
    ]

    for i, (a, b, expected_exception) in enumerate(cases):
        with subtests.test(msg=f"case_{i}"):
            with pytest.raises(expected_exception):
                a += b


def test_mask_merge_success(subtests) -> None:
    from nebius.base.fieldmask import Mask

    cases = [
        (Mask(), Mask(), ""),
        (Mask(any=Mask()), Mask(), "*"),
        (Mask(any=Mask()), Mask(any=Mask()), "*"),
        (
            Mask(any=Mask(any=Mask())),
            Mask(any=Mask()),
            "*.*",
        ),
        (
            Mask(any=Mask()),
            Mask(field_parts={"test": Mask()}),
            "*,test",
        ),
        (
            Mask(any=Mask()),
            Mask(field_parts={}),
            "*",
        ),
        (
            Mask(
                any=Mask(
                    any=Mask(
                        field_parts={
                            "a": Mask(),
                            "b": Mask(),
                            "c": Mask(any=Mask()),
                            "d": Mask(
                                field_parts={
                                    "e": Mask(),
                                    "f": Mask(),
                                }
                            ),
                            "g": Mask(
                                any=Mask(
                                    field_parts={
                                        "h": Mask(),
                                        "i": Mask(),
                                    }
                                ),
                                field_parts={
                                    "j": Mask(),
                                    "k": Mask(),
                                },
                            ),
                        }
                    )
                ),
                field_parts={
                    "l": Mask(),
                    "m": Mask(),
                    "n": Mask(any=Mask()),
                    "o": Mask(
                        field_parts={
                            "p": Mask(),
                            "q": Mask(),
                        }
                    ),
                    "r": Mask(
                        any=Mask(
                            field_parts={
                                "s": Mask(),
                                "t": Mask(),
                            }
                        ),
                        field_parts={
                            "u": Mask(any=Mask()),
                            "v": Mask(),
                        },
                    ),
                },
            ),
            Mask(
                any=Mask(
                    any=Mask(
                        field_parts={
                            "a": Mask(),
                            "b": Mask(),
                            "c": Mask(any=Mask()),
                            "d": Mask(
                                field_parts={
                                    "e": Mask(),
                                    "f": Mask(),
                                }
                            ),
                            "g": Mask(
                                any=Mask(
                                    field_parts={
                                        "h": Mask(),
                                        "i": Mask(),
                                    }
                                ),
                                field_parts={
                                    "j": Mask(),
                                    "k": Mask(),
                                },
                            ),
                        }
                    )
                ),
                field_parts={
                    "l": Mask(),
                    "A": Mask(),
                    "B": Mask(),
                    "1": Mask(),
                    "o": Mask(
                        field_parts={
                            "w": Mask(),
                            "x": Mask(),
                        }
                    ),
                    "z": Mask(
                        any=Mask(
                            field_parts={
                                "y": Mask(),
                            }
                        ),
                        field_parts={
                            "u": Mask(any=Mask()),
                            "v": Mask(),
                        },
                    ),
                },
            ),
            "*.*.(a,b,c.*,d.(e,f),g.(*.(h,i),j,k)),1,A,B,l,m,n.*,o.(p,q,w,x),r.(*.(s,t),u.*,v),z.(*.y,u.*,v)",
        ),
    ]

    for i, (a, b, expected_result) in enumerate(cases):
        with subtests.test(msg=f"case_{i}"):
            a += b
            result = a.marshal()
            assert result == expected_result


def test_to_field_path(subtests) -> None:
    from nebius.base.fieldmask import Error, FieldPath, Mask

    cases = [
        (Mask(any=Mask()), Error("wildcard in the mask")),
        (Mask(field_parts={"b": Mask(any=Mask())}), Error("wildcard in the mask")),
        (Mask(), None),
        (Mask(field_parts={}), None),
        (
            Mask(field_parts={"a": Mask(), "b": Mask()}),
            Error("multiple paths in the mask"),
        ),
        (
            Mask(field_parts={"c": Mask(field_parts={"a": Mask(), "b": Mask()})}),
            Error("multiple paths in the mask"),
        ),
        (
            Mask(field_parts={"a": Mask(field_parts={"b": Mask()})}),
            FieldPath(["a", "b"]),
        ),
        (
            Mask(field_parts={"a": Mask(field_parts={})}),
            FieldPath(["a"]),
        ),
    ]

    for i, (mask, expected) in enumerate(cases):
        with subtests.test(msg=f"case_{i}"):
            if isinstance(expected, Error):
                with pytest.raises(Error) as exc_info:
                    mask.to_field_path()
                assert str(exc_info.value) == str(expected)
                assert not mask.is_field_path()
            else:
                assert mask.to_field_path() == expected
                assert mask.is_field_path()


def test_mask_sub_mask(subtests):
    from nebius.base.fieldmask import FieldKey, Mask

    inf_mask = Mask()
    inf_mask.field_parts["x"] = inf_mask

    cases = [
        {"mask": Mask(), "key": "foo", "result": None, "err": None},
        {"mask": None, "key": "foo", "result": None, "err": None},
        {
            "mask": Mask(field_parts={"bar": Mask()}),
            "key": "foo",
            "result": None,
            "err": None,
        },  # noqa: E501
        {
            "mask": Mask(field_parts={"foo": Mask(field_parts={"bar": Mask()})}),
            "key": "foo",
            "result": Mask(field_parts={"bar": Mask()}),
            "err": None,
        },
        {
            "mask": Mask(any=Mask(field_parts={"bar": Mask()})),
            "key": "foo",
            "result": Mask(field_parts={"bar": Mask()}),
            "err": None,
        },
        {
            "mask": Mask(
                any=Mask(field_parts={"baz": Mask()}),
                field_parts={"foo": Mask(field_parts={"bar": Mask()})},
            ),
            "key": "foo",
            "result": Mask(field_parts={"bar": Mask(), "baz": Mask()}),
            "err": None,
        },
        {
            "mask": Mask(
                any=Mask(field_parts={"baz": inf_mask}),
                field_parts={"foo": Mask(field_parts={"bar": Mask()})},
            ),
            "key": "foo",
            "result": None,
            "err": RecursionError,
        },
    ]

    for i, case in enumerate(cases):
        with subtests.test(msg=f"case_{i}"):
            mask = case["mask"]
            key = FieldKey(case["key"])
            expected_result = case["result"]
            expected_error = case["err"]

            if expected_error:
                with pytest.raises(expected_error):
                    result = mask.sub_mask(key) if mask else None
            else:
                result = mask.sub_mask(key) if mask else None
                assert result == expected_result, f"Case {i} failed"


def test_mask_sub_mask_by_path(subtests):
    from nebius.base.fieldmask import FieldPath, Mask

    inf_mask = Mask()
    inf_mask.field_parts["x"] = inf_mask

    cases = [
        {"mask": Mask(), "path": FieldPath(["foo"]), "result": None, "err": None},
        {
            "mask": Mask(field_parts={"bar": Mask()}),
            "path": FieldPath(["foo"]),
            "result": None,
            "err": None,
        },  # noqa: E501
        {
            "mask": Mask(field_parts={"foo": Mask(field_parts={"bar": Mask()})}),
            "path": FieldPath(["foo"]),
            "result": Mask(field_parts={"bar": Mask()}),
            "err": None,
        },
        {
            "mask": Mask(any=Mask(field_parts={"bar": Mask()})),
            "path": FieldPath(["foo"]),
            "result": Mask(field_parts={"bar": Mask()}),
            "err": None,
        },
        {
            "mask": Mask(
                any=Mask(field_parts={"baz": Mask()}),
                field_parts={"foo": Mask(field_parts={"bar": Mask()})},
            ),
            "path": FieldPath(["foo"]),
            "result": Mask(field_parts={"bar": Mask(), "baz": Mask()}),
            "err": None,
        },
        {
            "mask": Mask(
                any=Mask(field_parts={"baz": inf_mask}),
                field_parts={"foo": Mask(field_parts={"bar": Mask()})},
            ),
            "path": FieldPath(["foo"]),
            "result": None,
            "err": RecursionError,
        },
    ]

    for i, case in enumerate(cases):
        with subtests.test(msg=f"case_{i}"):
            mask = case["mask"]
            path = case["path"]
            expected_result = case["result"]
            expected_error = case["err"]

            if expected_error:
                with pytest.raises(expected_error):
                    result = mask.sub_mask(path) if mask else None
            else:
                result = mask.sub_mask(path) if mask else None
                assert result == expected_result, f"Case {i} failed"


def test_mask_intesect_reset_mask(subtests):
    from nebius.base.fieldmask import Mask

    inf_mask = Mask()
    inf_mask.field_parts["x"] = inf_mask

    assert Mask().intersect_reset_mask(None) is None
    assert Mask().intersect_reset_mask("") is None
    assert Mask().intersect_reset_mask(object()) is None

    with pytest.raises(RecursionError):
        inf_mask.intersect_reset_mask(inf_mask)
    with pytest.raises(RecursionError):
        Mask(any=inf_mask).intersect_reset_mask(Mask(any=inf_mask))

    cases = [
        ("", "", ""),
        ("a.(x,y,z),b.*", "*.x,a.z", "a.(x,z),b.x"),
        ("a", "*", "a"),
    ]
    for i, case in enumerate(cases):
        with subtests.test(msg=f"case_{i}"):
            a, b, r = case
            m = Mask.unmarshal(a).intersect_reset_mask(Mask.unmarshal(b))
            assert m == Mask.unmarshal(r), f"Case {i} failed"


def test_mask_intesect_dumb(
    subtests: SubTests,
):
    from nebius.base.fieldmask import Mask

    inf_mask = Mask()
    inf_mask.field_parts["x"] = inf_mask

    assert Mask().intersect_dumb(None) is None
    assert Mask().intersect_dumb("") is None
    assert Mask().intersect_dumb(object()) is None

    with pytest.raises(RecursionError):
        inf_mask.intersect_dumb(inf_mask)
    with pytest.raises(RecursionError):
        Mask(any=inf_mask).intersect_dumb(Mask(any=inf_mask))

    cases = [
        ("", "", ""),
        ("*.(a,b),x,y", "*.(a,b),x,y", "*.(a,b),x,y"),
        ("a", "*", ""),
        ("*.(a,b),x,y", "*.(x,y),z,f", "*"),
        ("*.(a,b),x,y", "*.(a,y),x,f", "*.a,x"),
    ]
    for i, case in enumerate(cases):
        with subtests.test(msg=f"case_{i}"):
            a, b, r = case
            m = Mask.unmarshal(a).intersect_dumb(Mask.unmarshal(b))
            assert m == Mask.unmarshal(r), f"Case {i} failed"


def test_mask_subtract_dumb(subtests: SubTests) -> None:
    from nebius.base.fieldmask import Mask

    inf_mask = Mask()
    inf_mask.field_parts["x"] = inf_mask

    assert Mask().subtract_dumb(None) is None
    assert Mask().subtract_dumb("") is None
    assert Mask().subtract_dumb(object()) is None

    with pytest.raises(RecursionError):
        inf_mask.subtract_dumb(inf_mask)
    with pytest.raises(RecursionError):
        Mask(any=inf_mask).subtract_dumb(Mask(any=inf_mask))

    cases = [
        ("x.(a,b),*.(c,d),e,f", "x.(a,b),*.(c,d),e,f", ""),
        ("x.(a,b),*.(c,d),e,f", "x.(a),*.(c),e", "x.b,*.d,f"),
        ("a", "*", "a"),
        ("a", "a", ""),
        ("a", "b", "a"),
    ]
    for i, case in enumerate(cases):
        with subtests.test(msg=f"case_{i}"):
            a, b, r = case
            mask_a = Mask.unmarshal(a)
            mask_b = Mask.unmarshal(b)
            mask_r = Mask.unmarshal(r)
            assert isinstance(mask_a, Mask), f"Case {i} failed: {a} must be mask"
            assert isinstance(mask_b, Mask), f"Case {i} failed: {b} must be mask"
            assert isinstance(mask_r, Mask), f"Case {i} failed: {r} must be mask"
            mask_a.subtract_dumb(mask_b)
            assert mask_a == mask_r, f"Case {i} failed: {mask_a} != {mask_r}"


def test_mask_subtract_reset_mask(subtests: SubTests) -> None:
    from nebius.base.fieldmask import Mask

    inf_mask = Mask()
    inf_mask.field_parts["x"] = inf_mask

    assert Mask().subtract_reset_mask(None) is None
    assert Mask().subtract_reset_mask("") is None
    assert Mask().subtract_reset_mask(object()) is None

    with pytest.raises(RecursionError):
        inf_mask.subtract_reset_mask(inf_mask)
    with pytest.raises(RecursionError):
        Mask(any=inf_mask).subtract_reset_mask(Mask(any=inf_mask))

    cases = [
        ("x.(a,b),*.(c,d),e,f", "x.(a,b),*.(c,d),e,f", ""),
        ("x.(a,b),*.(c,d),e,f", "x.(a),*.(c),e", "x.b,*.d"),
    ]
    for i, case in enumerate(cases):
        with subtests.test(msg=f"case_{i}"):
            a, b, r = case
            mask_a = Mask.unmarshal(a)
            mask_b = Mask.unmarshal(b)
            mask_r = Mask.unmarshal(r)
            assert isinstance(mask_a, Mask), f"Case {i} failed: {a} must be mask"
            assert isinstance(mask_b, Mask), f"Case {i} failed: {b} must be mask"
            assert isinstance(mask_r, Mask), f"Case {i} failed: {r} must be mask"
            mask_a.subtract_reset_mask(mask_b)
            assert mask_a == mask_r, f"Case {i} failed: {mask_a} != {mask_r}"



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: nebius/iam/v2/tenant.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from nebius.api.buf.validate import validate_pb2 as buf_dot_validate_dot_validate__pb2
from nebius.api.nebius import annotations_pb2 as nebius_dot_annotations__pb2
from nebius.api.nebius.common.v1 import metadata_pb2 as nebius_dot_common_dot_v1_dot_metadata__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x1anebius/iam/v2/tenant.proto\x12\rnebius.iam.v2\x1a\x1b\x62uf/validate/validate.proto\x1a\x18nebius/annotations.proto\x1a\x1fnebius/common/v1/metadata.proto\"\xc2\x01\n\x06Tenant\x12\x46\n\x08metadata\x18\x01 \x01(\x0b\x32\".nebius.common.v1.ResourceMetadataB\x06\xbaH\x03\xc8\x01\x01R\x08metadata\x12\x35\n\x04spec\x18\x02 \x01(\x0b\x32\x19.nebius.iam.v2.TenantSpecB\x06\xbaH\x03\xc8\x01\x01R\x04spec\x12\x39\n\x06status\x18\x03 \x01(\x0b\x32\x1b.nebius.iam.v2.TenantStatusB\x04\xbaJ\x01\x05R\x06status\"\x0c\n\nTenantSpec\"\x9a\x01\n\x0cTenantStatus\x12J\n\x0ctenant_state\x18\x01 \x01(\x0e\x32\'.nebius.iam.v2.TenantStatus.TenantStateR\x0btenantState\">\n\x0bTenantState\x12\x15\n\x11STATE_UNSPECIFIED\x10\x00\x12\x0c\n\x08\x43REATING\x10\x01\x12\n\n\x06\x41\x43TIVE\x10\x02\x42R\n\x14\x61i.nebius.pub.iam.v2B\x0bTenantProtoP\x01Z+github.com/nebius/gosdk/proto/nebius/iam/v2b\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'nebius.iam.v2.tenant_pb2', _globals)
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\n\024ai.nebius.pub.iam.v2B\013TenantProtoP\001Z+github.com/nebius/gosdk/proto/nebius/iam/v2'
  _TENANT.fields_by_name['metadata']._options = None
  _TENANT.fields_by_name['metadata']._serialized_options = b'\272H\003\310\001\001'
  _TENANT.fields_by_name['spec']._options = None
  _TENANT.fields_by_name['spec']._serialized_options = b'\272H\003\310\001\001'
  _TENANT.fields_by_name['status']._options = None
  _TENANT.fields_by_name['status']._serialized_options = b'\272J\001\005'
  _globals['_TENANT']._serialized_start=134
  _globals['_TENANT']._serialized_end=328
  _globals['_TENANTSPEC']._serialized_start=330
  _globals['_TENANTSPEC']._serialized_end=342
  _globals['_TENANTSTATUS']._serialized_start=345
  _globals['_TENANTSTATUS']._serialized_end=499
  _globals['_TENANTSTATUS_TENANTSTATE']._serialized_start=437
  _globals['_TENANTSTATUS_TENANTSTATE']._serialized_end=499
# @@protoc_insertion_point(module_scope)



from google.protobuf.internal import enum_type_wrapper as _enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Optional as _Optional, Union as _Union

DESCRIPTOR: _descriptor.FileDescriptor

class ClusterStatus(_message.Message):
    __slots__ = ["phase", "state", "reconciling"]
    class Phase(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
        __slots__ = []
        PHASE_UNSPECIFIED: _ClassVar[ClusterStatus.Phase]
        PHASE_PROVISIONING: _ClassVar[ClusterStatus.Phase]
        PHASE_RUNNING: _ClassVar[ClusterStatus.Phase]
        PHASE_UPDATING: _ClassVar[ClusterStatus.Phase]
        PHASE_DELETING: _ClassVar[ClusterStatus.Phase]
        PHASE_DELETED: _ClassVar[ClusterStatus.Phase]
        PHASE_PURGING: _ClassVar[ClusterStatus.Phase]
        PHASE_STOPPING: _ClassVar[ClusterStatus.Phase]
        PHASE_RESUMING: _ClassVar[ClusterStatus.Phase]
    PHASE_UNSPECIFIED: ClusterStatus.Phase
    PHASE_PROVISIONING: ClusterStatus.Phase
    PHASE_RUNNING: ClusterStatus.Phase
    PHASE_UPDATING: ClusterStatus.Phase
    PHASE_DELETING: ClusterStatus.Phase
    PHASE_DELETED: ClusterStatus.Phase
    PHASE_PURGING: ClusterStatus.Phase
    PHASE_STOPPING: ClusterStatus.Phase
    PHASE_RESUMING: ClusterStatus.Phase
    class State(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
        __slots__ = []
        STATE_UNSPECIFIED: _ClassVar[ClusterStatus.State]
        STATE_IN_PROGRESS: _ClassVar[ClusterStatus.State]
        STATE_FINISHED: _ClassVar[ClusterStatus.State]
        STATE_ERROR: _ClassVar[ClusterStatus.State]
        STATE_DEGRADED: _ClassVar[ClusterStatus.State]
        STATE_SCHEDULED: _ClassVar[ClusterStatus.State]
    STATE_UNSPECIFIED: ClusterStatus.State
    STATE_IN_PROGRESS: ClusterStatus.State
    STATE_FINISHED: ClusterStatus.State
    STATE_ERROR: ClusterStatus.State
    STATE_DEGRADED: ClusterStatus.State
    STATE_SCHEDULED: ClusterStatus.State
    PHASE_FIELD_NUMBER: _ClassVar[int]
    STATE_FIELD_NUMBER: _ClassVar[int]
    RECONCILING_FIELD_NUMBER: _ClassVar[int]
    phase: ClusterStatus.Phase
    state: ClusterStatus.State
    reconciling: bool
    def __init__(self, phase: _Optional[_Union[ClusterStatus.Phase, str]] = ..., state: _Optional[_Union[ClusterStatus.State, str]] = ..., reconciling: bool = ...) -> None: ...



from nebius.api.buf.validate import validate_pb2 as _validate_pb2
from nebius.api.nebius.common.v1 import operation_pb2 as _operation_pb2
from google.protobuf.internal import containers as _containers
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Iterable as _Iterable, Mapping as _Mapping, Optional as _Optional, Union as _Union

DESCRIPTOR: _descriptor.FileDescriptor

class GetOperationRequest(_message.Message):
    __slots__ = ["id"]
    ID_FIELD_NUMBER: _ClassVar[int]
    id: str
    def __init__(self, id: _Optional[str] = ...) -> None: ...

class ListOperationsRequest(_message.Message):
    __slots__ = ["resource_id", "page_size", "page_token"]
    RESOURCE_ID_FIELD_NUMBER: _ClassVar[int]
    PAGE_SIZE_FIELD_NUMBER: _ClassVar[int]
    PAGE_TOKEN_FIELD_NUMBER: _ClassVar[int]
    resource_id: str
    page_size: int
    page_token: str
    def __init__(self, resource_id: _Optional[str] = ..., page_size: _Optional[int] = ..., page_token: _Optional[str] = ...) -> None: ...

class ListOperationsResponse(_message.Message):
    __slots__ = ["operations", "next_page_token"]
    OPERATIONS_FIELD_NUMBER: _ClassVar[int]
    NEXT_PAGE_TOKEN_FIELD_NUMBER: _ClassVar[int]
    operations: _containers.RepeatedCompositeFieldContainer[_operation_pb2.Operation]
    next_page_token: str
    def __init__(self, operations: _Optional[_Iterable[_Union[_operation_pb2.Operation, _Mapping]]] = ..., next_page_token: _Optional[str] = ...) -> None: ...



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: nebius/mysterybox/v1/secret_version.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from nebius.api.buf.validate import validate_pb2 as buf_dot_validate_dot_validate__pb2
from google.protobuf import timestamp_pb2 as google_dot_protobuf_dot_timestamp__pb2
from nebius.api.nebius.common.v1 import metadata_pb2 as nebius_dot_common_dot_v1_dot_metadata__pb2
from nebius.api.nebius import annotations_pb2 as nebius_dot_annotations__pb2
from nebius.api.nebius.mysterybox.v1 import payload_pb2 as nebius_dot_mysterybox_dot_v1_dot_payload__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n)nebius/mysterybox/v1/secret_version.proto\x12\x14nebius.mysterybox.v1\x1a\x1b\x62uf/validate/validate.proto\x1a\x1fgoogle/protobuf/timestamp.proto\x1a\x1fnebius/common/v1/metadata.proto\x1a\x18nebius/annotations.proto\x1a\"nebius/mysterybox/v1/payload.proto\"\xdb\x01\n\rSecretVersion\x12>\n\x08metadata\x18\x01 \x01(\x0b\x32\".nebius.common.v1.ResourceMetadataR\x08metadata\x12;\n\x04spec\x18\x02 \x01(\x0b\x32\'.nebius.mysterybox.v1.SecretVersionSpecR\x04spec\x12G\n\x06status\x18\x03 \x01(\x0b\x32).nebius.mysterybox.v1.SecretVersionStatusB\x04\xbaJ\x01\x05R\x06status:\x04\xbaJ\x01\x02\"\xa8\x01\n\x11SecretVersionSpec\x12&\n\x0b\x64\x65scription\x18\x01 \x01(\tB\x04\xbaJ\x01\x02R\x0b\x64\x65scription\x12\x44\n\x07payload\x18\x02 \x03(\x0b\x32\x1d.nebius.mysterybox.v1.PayloadB\x0b\xbaH\x03\xc8\x01\x01\xbaJ\x02\x04\x02R\x07payload\x12%\n\x0bset_primary\x18\x03 \x01(\x08\x42\x04\xbaJ\x01\x04R\nsetPrimary\"\x96\x02\n\x13SecretVersionStatus\x12\x45\n\x05state\x18\x01 \x01(\x0e\x32/.nebius.mysterybox.v1.SecretVersionStatus.StateR\x05state\x12\x39\n\ndeleted_at\x18\x02 \x01(\x0b\x32\x1a.google.protobuf.TimestampR\tdeletedAt\x12\x35\n\x08purge_at\x18\x03 \x01(\x0b\x32\x1a.google.protobuf.TimestampR\x07purgeAt\"F\n\x05State\x12\x15\n\x11STATE_UNSPECIFIED\x10\x00\x12\n\n\x06\x41\x43TIVE\x10\x01\x12\x1a\n\x16SCHEDULED_FOR_DELETION\x10\x02\x42g\n\x1b\x61i.nebius.pub.mysterybox.v1B\x12SecretVersionProtoP\x01Z2github.com/nebius/gosdk/proto/nebius/mysterybox/v1b\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'nebius.mysterybox.v1.secret_version_pb2', _globals)
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\n\033ai.nebius.pub.mysterybox.v1B\022SecretVersionProtoP\001Z2github.com/nebius/gosdk/proto/nebius/mysterybox/v1'
  _SECRETVERSION.fields_by_name['status']._options = None
  _SECRETVERSION.fields_by_name['status']._serialized_options = b'\272J\001\005'
  _SECRETVERSION._options = None
  _SECRETVERSION._serialized_options = b'\272J\001\002'
  _SECRETVERSIONSPEC.fields_by_name['description']._options = None
  _SECRETVERSIONSPEC.fields_by_name['description']._serialized_options = b'\272J\001\002'
  _SECRETVERSIONSPEC.fields_by_name['payload']._options = None
  _SECRETVERSIONSPEC.fields_by_name['payload']._serialized_options = b'\272H\003\310\001\001\272J\002\004\002'
  _SECRETVERSIONSPEC.fields_by_name['set_primary']._options = None
  _SECRETVERSIONSPEC.fields_by_name['set_primary']._serialized_options = b'\272J\001\004'
  _globals['_SECRETVERSION']._serialized_start=225
  _globals['_SECRETVERSION']._serialized_end=444
  _globals['_SECRETVERSIONSPEC']._serialized_start=447
  _globals['_SECRETVERSIONSPEC']._serialized_end=615
  _globals['_SECRETVERSIONSTATUS']._serialized_start=618
  _globals['_SECRETVERSIONSTATUS']._serialized_end=896
  _globals['_SECRETVERSIONSTATUS_STATE']._serialized_start=826
  _globals['_SECRETVERSIONSTATUS_STATE']._serialized_end=896
# @@protoc_insertion_point(module_scope)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: nebius/iam/v1/access_key.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from nebius.api.buf.validate import validate_pb2 as buf_dot_validate_dot_validate__pb2
from google.protobuf import timestamp_pb2 as google_dot_protobuf_dot_timestamp__pb2
from nebius.api.nebius.common.v1 import metadata_pb2 as nebius_dot_common_dot_v1_dot_metadata__pb2
from nebius.api.nebius import annotations_pb2 as nebius_dot_annotations__pb2
from nebius.api.nebius.iam.v1 import access_pb2 as nebius_dot_iam_dot_v1_dot_access__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x1enebius/iam/v1/access_key.proto\x12\rnebius.iam.v1\x1a\x1b\x62uf/validate/validate.proto\x1a\x1fgoogle/protobuf/timestamp.proto\x1a\x1fnebius/common/v1/metadata.proto\x1a\x18nebius/annotations.proto\x1a\x1anebius/iam/v1/access.proto\"\xd1\x01\n\tAccessKey\x12\x46\n\x08metadata\x18\x01 \x01(\x0b\x32\".nebius.common.v1.ResourceMetadataB\x06\xbaH\x03\xc8\x01\x01R\x08metadata\x12\x38\n\x04spec\x18\x02 \x01(\x0b\x32\x1c.nebius.iam.v1.AccessKeySpecB\x06\xbaH\x03\xc8\x01\x01R\x04spec\x12<\n\x06status\x18\x03 \x01(\x0b\x32\x1e.nebius.iam.v1.AccessKeyStatusB\x04\xbaJ\x01\x05R\x06status:\x04\xbaJ\x01\x02\"\xaa\x01\n\rAccessKeySpec\x12\x36\n\x07\x61\x63\x63ount\x18\x01 \x01(\x0b\x32\x16.nebius.iam.v1.AccountB\x04\xbaJ\x01\x02R\x07\x61\x63\x63ount\x12?\n\nexpires_at\x18\x02 \x01(\x0b\x32\x1a.google.protobuf.TimestampB\x04\xbaJ\x01\x02R\texpiresAt\x12 \n\x0b\x64\x65scription\x18\x03 \x01(\tR\x0b\x64\x65scription\"\xd2\x02\n\x0f\x41\x63\x63\x65ssKeyStatus\x12:\n\x05state\x18\x01 \x01(\x0e\x32$.nebius.iam.v1.AccessKeyStatus.StateR\x05state\x12 \n\x0b\x66ingerprint\x18\x02 \x01(\tR\x0b\x66ingerprint\x12\x1c\n\talgorithm\x18\x03 \x01(\tR\talgorithm\x12\x19\n\x08key_size\x18\x04 \x01(\x05R\x07keySize\x12)\n\x11\x61ws_access_key_id\x18\x05 \x01(\tR\x0e\x61wsAccessKeyId\x12\x1b\n\x06secret\x18\x06 \x01(\tB\x03\xc0J\x01R\x06secret\"`\n\x05State\x12\x15\n\x11STATE_UNSPECIFIED\x10\x00\x12\n\n\x06\x41\x43TIVE\x10\x01\x12\x0c\n\x08INACTIVE\x10\x02\x12\x0b\n\x07\x45XPIRED\x10\x03\x12\x0c\n\x08\x44\x45LETING\x10\x04\x12\x0b\n\x07\x44\x45LETED\x10\x05\x42U\n\x14\x61i.nebius.pub.iam.v1B\x0e\x41\x63\x63\x65ssKeyProtoP\x01Z+github.com/nebius/gosdk/proto/nebius/iam/v1b\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'nebius.iam.v1.access_key_pb2', _globals)
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\n\024ai.nebius.pub.iam.v1B\016AccessKeyProtoP\001Z+github.com/nebius/gosdk/proto/nebius/iam/v1'
  _ACCESSKEY.fields_by_name['metadata']._options = None
  _ACCESSKEY.fields_by_name['metadata']._serialized_options = b'\272H\003\310\001\001'
  _ACCESSKEY.fields_by_name['spec']._options = None
  _ACCESSKEY.fields_by_name['spec']._serialized_options = b'\272H\003\310\001\001'
  _ACCESSKEY.fields_by_name['status']._options = None
  _ACCESSKEY.fields_by_name['status']._serialized_options = b'\272J\001\005'
  _ACCESSKEY._options = None
  _ACCESSKEY._serialized_options = b'\272J\001\002'
  _ACCESSKEYSPEC.fields_by_name['account']._options = None
  _ACCESSKEYSPEC.fields_by_name['account']._serialized_options = b'\272J\001\002'
  _ACCESSKEYSPEC.fields_by_name['expires_at']._options = None
  _ACCESSKEYSPEC.fields_by_name['expires_at']._serialized_options = b'\272J\001\002'
  _ACCESSKEYSTATUS.fields_by_name['secret']._options = None
  _ACCESSKEYSTATUS.fields_by_name['secret']._serialized_options = b'\300J\001'
  _globals['_ACCESSKEY']._serialized_start=199
  _globals['_ACCESSKEY']._serialized_end=408
  _globals['_ACCESSKEYSPEC']._serialized_start=411
  _globals['_ACCESSKEYSPEC']._serialized_end=581
  _globals['_ACCESSKEYSTATUS']._serialized_start=584
  _globals['_ACCESSKEYSTATUS']._serialized_end=922
  _globals['_ACCESSKEYSTATUS_STATE']._serialized_start=826
  _globals['_ACCESSKEYSTATUS_STATE']._serialized_end=922
# @@protoc_insertion_point(module_scope)



"""Asynchronous renewable file-backed bearer.

This module provides an asynchronous variant of the renewable file-backed
bearer which uses a :class:`ThrottledTokenCache` and runs a background
refresh task. The implementation exposes two primary classes:

- :class:`AsynchronousRenewableFileCacheReceiver` -- An async receiver that
    prefers the cached token when it's fresh and, when required, requests a
    renewal from the wrapped bearer. It implements a limited retry strategy.
- :class:`AsynchronousRenewableFileCacheBearer` -- A bearer that wraps an
    existing :class:`nebius.aio.token.token.Bearer` and manages a background
    refresh loop that persists tokens to a file-backed cache.

Example
-------

Create an asynchronous renewable bearer from a named network bearer::

        async_bearer = AsynchronousRenewableFileCacheBearer(network_bearer)
        token = await async_bearer.receiver().fetch()

"""

from asyncio import (
    FIRST_COMPLETED,
    CancelledError,
    Event,
    Future,
    Task,
    create_task,
    gather,
    sleep,
    wait,
    wait_for,
)
from collections.abc import Awaitable
from datetime import datetime, timedelta, timezone
from logging import getLogger
from pathlib import Path
from typing import Any, TypeVar

from nebius.aio.token.token import Bearer as ParentBearer
from nebius.aio.token.token import Receiver as ParentReceiver
from nebius.aio.token.token import Token
from nebius.base.constants import DEFAULT_CONFIG_DIR, DEFAULT_CREDENTIALS_FILE

from ..options import (
    OPTION_MAX_RETRIES,
    OPTION_RENEW_REQUEST_TIMEOUT,
    OPTION_RENEW_REQUIRED,
    OPTION_RENEW_SYNCHRONOUS,
    OPTION_REPORT_ERROR,
)
from ..renewable import (
    RenewalError,
)
from .throttled_token_cache import ThrottledTokenCache

log = getLogger(__name__)


class AsynchronousRenewableFileCacheReceiver(ParentReceiver):
    """Receiver that requests asynchronous renewals when needed.

    The receiver delegates to the parent :class:`AsynchronousRenewableFileCacheBearer`
    for actual fetching and renewal coordination. It tracks a small retry
    counter so transient errors can be retried a configurable number of times.

    Example
    -------

    Constructing a receiver is normally done via the bearer's
    :meth:`AsynchronousRenewableFileCacheBearer.receiver` method::

        receiver = async_bearer.receiver()

    :param bearer: The owning :class:`AsynchronousRenewableFileCacheBearer`.
    :param max_retries: Maximum number of automatic retry attempts before
        giving up.
    """

    def __init__(
        self,
        bearer: "AsynchronousRenewableFileCacheBearer",
        max_retries: int = 2,
    ) -> None:
        """Initialize the receiver.

        No I/O is performed during construction.
        """
        super().__init__()
        self._parent = bearer
        self._max_retries = max_retries
        self._trial = 0

    async def _fetch(
        self, timeout: float | None = None, options: dict[str, str] | None = None
    ) -> Token:
        """Fetch a token, forwarding to the parent bearer.

        The method increments an internal trial count used by
        :meth:`can_retry` and delegates the actual fetch to the parent
        bearer's :meth:`AsynchronousRenewableFileCacheBearer.fetch` method.

        :param timeout: Optional timeout forwarded to the parent fetch.
        :param options: Optional fetch options forwarded to the parent.
        :returns: A :class:`Token` instance.
        """
        self._trial += 1
        log.debug(
            f"token fetch requested, attempt: {self._trial}," f"timeout: {timeout}"
        )
        return await self._parent.fetch(timeout=timeout, options=options)

    def can_retry(
        self,
        err: Exception,
        options: dict[str, str] | None = None,
    ) -> bool:
        """Decide whether the caller should retry the operation.

        The receiver inspects provided options and the internal trial count to
        decide whether to allow another retry. When retries are permitted and
        the operation was not requested synchronously the receiver will
        proactively request a background renewal via the parent bearer.

        :param err: The exception that occurred during fetch.
        :param options: Optional map of string options. Supported keys include
            :data:`OPTION_RENEW_SYNCHRONOUS` and :data:`OPTION_MAX_RETRIES` which can
            override synchronous behavior and the max retry count.
        :returns: `True` if the caller should retry, `False` otherwise.
        """
        max_retries = self._max_retries
        synchronous = False
        if options is not None:
            synchronous = options.get(OPTION_RENEW_SYNCHRONOUS, "") != ""
            if OPTION_MAX_RETRIES in options:
                value = options[OPTION_MAX_RETRIES]
                try:
                    max_retries = int(value)
                except ValueError as val_err:
                    log.error(
                        f"option {OPTION_MAX_RETRIES} is not valid integer: {val_err=}"
                    )
        if self._trial >= max_retries:
            log.debug("max retries reached, cannot retry")
            return False
        if not synchronous:
            # Proactively request an asynchronous renewal so other callers
            # may receive a fresh token.
            self._parent.request_renewal()
        return True


VERY_LONG_TIMEOUT = timedelta(days=365 * 10)  # 10 years, should be enough for anyone

"""A very long timeout used when a token has no expiration.

This constant is used as an effectively infinite timeout value when
calculating retry intervals for tokens that carry no expiration time.
"""

T = TypeVar("T")


class AsynchronousRenewableFileCacheBearer(ParentBearer):
    """Asynchronous bearer that refreshes tokens into a file-backed cache.

    This bearer wraps a network-facing :class:`nebius.aio.token.token.Bearer`
    and uses a :class:`ThrottledTokenCache` for persistence. A background
    task may be spawned to proactively refresh the token before expiry. The
    class exposes control methods such as :meth:`request_renewal`,
    :meth:`stop` and :meth:`close` to coordinate lifecycle and shutdown.

    :ivar safety_margin: A :class:`datetime.timedelta` or seconds value used to
        determine when a token is considered too close to expiry. When ``None``
        the cache will be used until tokens are actually expired.

    :param source: Wrapped bearer used to obtain fresh tokens. The source
        must have a non-empty :attr:`nebius.aio.token.token.Bearer.name` used as the
        cache key.
    :param max_retries: Default maximum retries for receivers produced by
        :meth:`receiver`.
    :param initial_safety_margin: Safety margin before token expiration used
        for the first fetch. The token will be immediately refreshed if the
        cache token is closer to expiry than this margin.
    :param retry_safety_margin: Safety margin applied when computing the
        initial retry interval for the background loop.
    :param lifetime_safe_fraction: Fraction of the token lifetime used to
        schedule proactive refreshes (e.g. 0.9 means refresh at 90% of life).
    :param initial_retry_timeout: Initial retry backoff timeout used after failures.
    :param max_retry_timeout: Maximum retry backoff timeout used when backing off.
    :param retry_timeout_exponent: Exponent used for exponential backoff.
    :param refresh_request_timeout: Default timeout forwarded to the
        wrapped receiver during refresh operations.
    :param file_cache_throttle: Throttle interval passed to
        :class:`ThrottledTokenCache` to reduce disk reads.

    Example
    -------

    Wrap a custom bearer with a name and file cache::

        from nebius.sdk import SDK
        from nebius.aio.token.token import NamedBearer, Bearer, Receiver, Token
        from nebius.aio.token.file_cache.async_renewable_bearer import (
            AsynchronousRenewableFileCacheBearer
        )

        class SomeCustomHeavyLoadBearer(Bearer):
            def receiver(self) -> Receiver:
                return SomeReceiver()

        class SomeReceiver(Receiver):
            async def _fetch(self, timeout=None, options=None) -> Token:
                # Simulate heavy load token fetch
                return Token("heavy-token")

            def can_retry(self, err, options=None) -> bool:
                return False

        custom_bearer = SomeCustomHeavyLoadBearer()
        named_bearer = NamedBearer(custom_bearer, "heavy-load-bearer")
        cached_bearer = AsynchronousRenewableFileCacheBearer(named_bearer)

        sdk = SDK(credentials=cached_bearer)
    """

    def __init__(
        self,
        source: ParentBearer,
        max_retries: int = 2,
        initial_safety_margin: timedelta | float | None = timedelta(hours=2),
        retry_safety_margin: timedelta = timedelta(hours=2),
        lifetime_safe_fraction: float = 0.9,
        initial_retry_timeout: timedelta = timedelta(seconds=1),
        max_retry_timeout: timedelta = timedelta(minutes=1),
        retry_timeout_exponent: float = 1.5,
        refresh_request_timeout: timedelta = timedelta(seconds=5),
        file_cache_throttle: timedelta | float = timedelta(minutes=5),
    ) -> None:
        """Initialize the asynchronous renewable bearer.

        The constructor constructs a :class:`ThrottledTokenCache` using the
        source bearer's :attr:`nebius.aio.token.token.Bearer.name` and prepares internal
        synchronization primitives used by the background refresh loop.

        :raises ValueError: When the wrapped bearer has no name.
        """
        super().__init__()
        self._source = source
        if isinstance(initial_safety_margin, (float, int)):
            initial_safety_margin = timedelta(seconds=initial_safety_margin)
        self._retry_safety_margin = retry_safety_margin
        self.safety_margin = initial_safety_margin
        name = self._source.name
        if name is None:
            raise ValueError("Source bearer must have a name for the cache.")
        self._file_cache = ThrottledTokenCache(
            name=name,
            cache_file=Path(DEFAULT_CONFIG_DIR) / DEFAULT_CREDENTIALS_FILE,
            throttle=file_cache_throttle,
        )

        # Synchronization and lifecycle events used by the refresh loop.
        self._is_fresh = Event()
        self._is_stopped = Event()
        self._renew_requested = Event()
        self._synchronous_can_proceed = Event()
        self._break_previous_attempt = Event()

        self._synchronous_can_proceed.set()

        # Background task handles and bookkeeping.
        self._refresh_task: Task[Any] | None = None
        self._tasks = set[Task[Any]]()

        self._renewal_attempt = 0

        self._max_retries = max_retries
        self._lifetime_safe_fraction = lifetime_safe_fraction
        self._initial_retry_timeout = initial_retry_timeout
        self._max_retry_timeout = max_retry_timeout
        self._retry_timeout_exponent = retry_timeout_exponent
        self._refresh_request_timeout = refresh_request_timeout

        # Fields used when synchronous renewal requests are made by callers.
        self._renew_synchronous_timeout: float | None = None
        self._renewal_future: Future[Token] | None = None
        self._renew_synchronous_options: dict[str, str] | None = None

    @property
    def wrapped(self) -> ParentBearer | None:
        """Return the wrapped bearer instance.

        :returns: The underlying bearer provided at construction.
        """
        return self._source

    def bg_task(self, coro: Awaitable[T]) -> Task[None]:
        """Run a coroutine in fire-and-forget mode.

        The helper wraps the provided coroutine in a simple error-handling
        wrapper so exceptions are logged rather than lost. The returned
        :class:`asyncio.Task` is tracked in :attr:`_tasks` and will be
        automatically removed when done.

        :param coro: Awaitable to schedule.
        :returns: The created :class:`asyncio.Task`.
        """

        async def wrapper() -> None:
            try:
                await coro
            except CancelledError:
                pass
            except Exception as e:
                log.error("Unhandled exception in fire-and-forget task", exc_info=e)

        ret = create_task(wrapper())
        ret.add_done_callback(lambda x: self._tasks.discard(x))
        self._tasks.add(ret)
        return ret

    async def fetch(
        self, timeout: float | None = None, options: dict[str, str] | None = None
    ) -> Token:
        """Fetch a token, using the file cache and optionally requesting renewal.

        The method attempts to return a cached token when it is considered
        fresh according to :attr:`safety_margin`. If the token is missing or
        too close to expiry a renewal is requested. The renewal may be handled
        asynchronously by the background refresh loop or synchronously if the
        caller requests it via provided options.

        Supported options
        -----------------
        - :data:`OPTION_RENEW_REQUIRED` -- Force a renewal even if a cached token
          exists.
        - :data:`OPTION_RENEW_SYNCHRONOUS` -- Request the renewal synchronously and
          wait for the result.
        - :data:`OPTION_REPORT_ERROR` -- Return exceptions from the renewal to the
          caller via the returned future.
        - :data:`OPTION_RENEW_REQUEST_TIMEOUT` -- When performing a synchronous
          request, this value (string float) can override the request timeout.

        :param timeout: Timeout in seconds for synchronous waits.
        :param options: Optional dictionary of string options (see above).
        :returns: A valid :class:`Token`.
        :raises RenewalError: If the cache is still empty after a renewal.
        """
        required = False
        synchronous = False
        report_error = False
        if options is not None:
            required = options.get(OPTION_RENEW_REQUIRED, "") != ""
            synchronous = options.get(OPTION_RENEW_SYNCHRONOUS, "") != ""
            report_error = options.get(OPTION_REPORT_ERROR, "") != ""

        tok = await self._file_cache.get()
        if not required and tok is not None and not tok.is_expired():
            if self.safety_margin is None or (
                not tok.expiration
                or (tok.expiration - self.safety_margin > datetime.now(timezone.utc))
            ):
                log.debug(f"token is fresh: {tok}")
                if self._refresh_task is None:
                    log.debug("no refresh task yet, starting it")
                    self._refresh_task = self.bg_task(self._run(True))
                return tok
        self.safety_margin = None  # reset safety margin after first fetch

        if self._refresh_task is None:
            log.debug("no refresh task yet, starting it")
            self._refresh_task = self.bg_task(self._run())
        if self.is_renewal_required() or required:
            log.debug(f"renewal required, timeout {timeout}")
            if synchronous:
                self._break_previous_attempt.set()
                await wait_for(self._synchronous_can_proceed.wait(), timeout)
                if OPTION_RENEW_REQUEST_TIMEOUT in options:  # type: ignore
                    try:
                        self._renew_synchronous_timeout = float(
                            options[OPTION_RENEW_REQUEST_TIMEOUT]  # type: ignore
                        )
                    except ValueError as err:
                        log.error(
                            f"option {OPTION_RENEW_REQUEST_TIMEOUT} value is not float:"
                            f" {err=}"
                        )
                self._renew_synchronous_options = options.copy()  # type: ignore
            if report_error or synchronous:
                self._renewal_future = Future[Token]()

            self._renew_requested.set()
            if report_error or synchronous:
                return await wait_for(self._renewal_future, timeout)  # type: ignore
            else:
                await wait_for(self._is_fresh.wait(), timeout)
        tok = await self._file_cache.get()
        if tok is None:
            raise RenewalError("cache is empty after renewal")
        return tok

    async def _fetch_once(self) -> Token:
        """Perform a single token refresh attempt.

        This helper arranges concurrent tasks for fetching a token from the
        wrapped bearer and for canceling that fetch if a newer synchronous
        request arrives. On success the new token is persisted into the
        file cache and any waiting synchronous future is fulfilled.

        :returns: The newly fetched :class:`Token`.
        """
        tok = None
        log.debug(f"refreshing token, attempt {self._renewal_attempt}")
        self._break_previous_attempt.clear()
        self._synchronous_can_proceed.clear()
        timeout = self._refresh_request_timeout.total_seconds()
        if self._renew_synchronous_timeout is not None:
            timeout = self._renew_synchronous_timeout
            self._renew_synchronous_timeout = None
        token_task = create_task(self._source.receiver().fetch(timeout))
        breaker_task = create_task(self._break_previous_attempt.wait())
        _done, pending = await wait(
            [token_task, breaker_task],
            return_when=FIRST_COMPLETED,
        )
        self._renew_requested.clear()
        self._synchronous_can_proceed.set()
        for t in pending:
            t.cancel()
            self.bg_task(t)
        tok = token_task.result()
        log.debug(f"received new token: {tok}")
        if self._renewal_future is not None and not self._renewal_future.done():
            self._renewal_future.set_result(tok)
        await self._file_cache.set(tok)
        self._renewal_attempt = 0
        self._is_fresh.set()
        return tok

    async def _run(self, wait_for_timeout: bool = False) -> None:
        """Background refresh loop.

        The loop waits either for a renewal request event or for a timer
        indicating the next refresh time. On renewal it calls
        :meth:`_fetch_once` and updates the interval used for the next run.
        Failures are logged and the retry interval is increased using an
        exponential backoff strategy up to ``_max_retry_timeout``.

        :param wait_for_timeout: When true the loop will compute an initial wait
            interval based on the currently cached token to avoid immediate
            refreshes.
        """
        log.debug("refresh task started")

        if wait_for_timeout:
            tok = await self._file_cache.get()
            if tok is not None and not tok.is_expired():
                if tok.expiration is not None:
                    retry_timeout = (
                        tok.expiration - datetime.now(timezone.utc)
                    ).total_seconds() - self._retry_safety_margin.total_seconds()
                else:
                    retry_timeout = VERY_LONG_TIMEOUT.total_seconds()
            else:
                retry_timeout = 0
        else:
            retry_timeout = 0

        while not self._is_stopped.is_set():
            log.debug(
                f"Will refresh token after {retry_timeout} seconds, "
                f"renewal attempt number {self._renewal_attempt}"
            )
            _done, pending = await wait(
                [
                    self.bg_task(self._renew_requested.wait()),
                    self.bg_task(sleep(retry_timeout)),
                ],
                return_when=FIRST_COMPLETED,
            )
            for task in pending:
                task.cancel()
            await gather(*pending, return_exceptions=True)
            try:
                tok = await self._fetch_once()
                exp = tok.expiration
                if exp is None:
                    retry_timeout = VERY_LONG_TIMEOUT.total_seconds()
                else:
                    retry_timeout = (
                        exp - datetime.now(timezone.utc)
                    ).total_seconds() * self._lifetime_safe_fraction
            except Exception as e:
                log.error(
                    f"Failed refresh token, attempt: {self._renewal_attempt}, "
                    f"error: {e}",
                    exc_info=e,
                )
                if self._renewal_future is not None and not self._renewal_future.done():
                    self._renewal_future.set_exception(e)
                if (
                    self._renewal_attempt <= 1
                    or abs(self._retry_timeout_exponent - 1) < 1e-9
                ):
                    retry_timeout = self._initial_retry_timeout.total_seconds()
                else:
                    mul = self._retry_timeout_exponent ** (self._renewal_attempt - 1)
                    retry_timeout = min(
                        self._initial_retry_timeout.total_seconds() * mul,
                        self._max_retry_timeout.total_seconds(),
                    )
            if retry_timeout < self._initial_retry_timeout.total_seconds():
                retry_timeout = self._initial_retry_timeout.total_seconds()

    async def close(self, grace: float | None = None) -> None:
        """Close the bearer and shutdown background tasks.

        The method attempts a graceful shutdown by requesting the wrapped
        source to close, signalling the background loop to stop and cancelling
        any outstanding fire-and-forget tasks. Non-cancellation exceptions
        raised during shutdown are logged.

        :param grace: Optional grace period forwarded to the wrapped
            bearer's :meth:`close` method.
        """
        source_close = create_task(self._source.close(grace=grace))
        self.stop()
        for task in self._tasks:
            task.cancel()
        rets = await gather(
            source_close,
            *self._tasks,
            return_exceptions=True,
        )
        for ret in rets:
            if isinstance(ret, BaseException) and not isinstance(ret, CancelledError):
                log.error(f"Error while graceful shutdown: {ret}", exc_info=ret)

    def is_renewal_required(self) -> bool:
        """Return whether a renewal is required.

        A renewal is considered required when no token is available in the
        in-memory cache or when a renewal request event has been set.
        """
        return self._file_cache.get_cached() is None or self._renew_requested.is_set()

    def request_renewal(self) -> None:
        """Request a background token renewal.

        This method sets internal events so the background loop will attempt a
        refresh. If the bearer has been stopped the request is ignored.
        """
        if not self._is_stopped.is_set():
            log.debug("token renewal requested")
            self._is_fresh.clear()
            self._renew_requested.set()

    def stop(self) -> None:
        """Stop the background refresh loop and break any running attempt.

        After calling this method the bearer will not perform further
        renewals until a new instance is created. The method also clears
        fresh-state events so callers waiting for tokens will fail fast.
        """
        log.debug("stopping renewal task")
        self._is_stopped.set()
        self._is_fresh.clear()
        self._break_previous_attempt.set()
        self._renew_requested.set()

    def receiver(self) -> AsynchronousRenewableFileCacheReceiver:
        """Return a new :class:`AsynchronousRenewableFileCacheReceiver`.

        :returns: A receiver bound to this bearer which will forward fetch
            requests to the bearer and participate in the retry semantics.
        """
        return AsynchronousRenewableFileCacheReceiver(
            self,
            max_retries=self._max_retries,
        )



import asyncio
import ssl
import sys
import urllib.parse
import webbrowser
from logging import getLogger
from typing import TextIO

import aiohttp
from attr import dataclass

from nebius.base.tls_certificates import get_system_certificates

from .constants import AUTH_ENDPOINT, TOKEN_ENDPOINT
from .is_wsl import is_wsl
from .pkce import PKCE
from .server import CallbackHandler

log = getLogger(__name__)


@dataclass
class GetTokenResult:
    access_token: str
    expires_in: int | None = None


def https_url(raw_url: str) -> str:
    if not raw_url.startswith(("http://", "https://")):
        return f"https://{raw_url}"
    return raw_url


async def open_browser(url: str) -> None:
    if sys.platform.startswith("linux") and is_wsl():
        import subprocess

        subprocess.run(  # noqa: S603
            ["cmd.exe", "/c", "start", url.replace("&", "^&")],  # noqa: S607
            check=True,
        )
    else:
        webbrowser.open(url)
    return None


async def get_code(
    client_id: str,
    auth_endpoint: str,
    federation_id: str,
    pkce_code: PKCE,
    writer: TextIO | None = None,
    no_browser_open: bool = False,
    timeout: float | None = 300,
) -> tuple[str, str]:
    auth_url = urllib.parse.urlparse(auth_endpoint)
    callback = CallbackHandler()
    await callback.listen_and_serve()

    redirect_uri = callback.addr

    params = {
        "response_type": "code",
        "client_id": client_id,
        "scope": "openid",
        "state": callback.state,
        "redirect_uri": redirect_uri,
        "code_challenge": pkce_code.challenge,
        "code_challenge_method": pkce_code.method,
    }
    if federation_id:
        params["federation-id"] = federation_id

    query = urllib.parse.urlencode(params)
    full_auth_url = auth_url._replace(query=query).geturl()

    if no_browser_open:
        msg = (
            "To complete the authentication process, open this link in your"
            f" browser:\n{full_auth_url}"
        )
        log.debug(
            "Browser won't be opened. Show link to user.", extra={"url": full_auth_url}
        )
        if writer:
            print(msg, file=writer)
        else:
            log.info(msg)
    else:
        msg = (
            "Switch to your browser to complete login. If it didn't open "
            f"automatically, use:\n{full_auth_url}"
        )
        log.debug("Attempting to open browser.", extra={"url": full_auth_url})
        if writer:
            print(msg, file=writer)
        else:
            log.info(msg)
        try:
            await open_browser(full_auth_url)
        except Exception as browser_err:
            await callback.shutdown()
            raise RuntimeError(f"Failed to open browser: {browser_err}")

    try:
        code = await asyncio.wait_for(callback.wait_for_code(), timeout=timeout)
        if not code:
            raise RuntimeError("No code received from the callback")
        return code, redirect_uri
    except asyncio.TimeoutError:
        raise TimeoutError("Timeout waiting for auth code")
    finally:
        await callback.shutdown()


async def get_token(
    client_id: str,
    token_url: str,
    code: str,
    redirect_uri: str,
    verifier: str,
    ssl_ctx: ssl.SSLContext | None = None,
) -> GetTokenResult:
    data = {
        "grant_type": "authorization_code",
        "code": code,
        "redirect_uri": redirect_uri,
        "client_id": client_id,
        "code_verifier": verifier,
    }
    if ssl_ctx is None:
        root_ca = get_system_certificates()
        ssl_ctx = ssl.create_default_context(cafile=root_ca)

    async with aiohttp.ClientSession() as session:
        async with session.post(
            token_url,
            data=data,
            headers={"Content-Type": "application/x-www-form-urlencoded"},
            ssl=ssl_ctx,
        ) as resp:
            if resp.status != 200:
                body = await resp.text()
                raise RuntimeError(f"Token request failed: {resp.status} {body}")
            ret = await resp.json()
            if not isinstance(ret, dict):
                raise RuntimeError(f"Unexpected response format: {ret}, expected dict")
            tok = ret.get("access_token", "")  # type: ignore[unused-ignore]
            expires_in = ret.get("expires_in", None)  # type: ignore[unused-ignore]
            if not isinstance(tok, str):
                raise RuntimeError(
                    f"Invalid token response: {ret}, expected 'access_token' as str"
                )
            if expires_in is not None and not isinstance(expires_in, int):
                raise RuntimeError(
                    f"Invalid token response: {ret}, expected 'expires_in' as int or"
                    " None"
                )
            return GetTokenResult(access_token=tok, expires_in=expires_in)


async def authorize(
    client_id: str,
    federation_endpoint: str,
    federation_id: str,
    writer: TextIO | None = None,
    no_browser_open: bool = False,
    timeout: float | None = 300,
    ssl_ctx: ssl.SSLContext | None = None,
) -> GetTokenResult:
    token_url = urllib.parse.urljoin(https_url(federation_endpoint), TOKEN_ENDPOINT)
    auth_url = urllib.parse.urljoin(https_url(federation_endpoint), AUTH_ENDPOINT)

    pkce = PKCE()

    code, redirect_uri = await get_code(
        client_id=client_id,
        auth_endpoint=auth_url,
        federation_id=federation_id,
        pkce_code=pkce,
        writer=writer,
        no_browser_open=no_browser_open,
        timeout=timeout,
    )

    log.debug("Auth code received")

    token = await get_token(
        client_id=client_id,
        token_url=token_url,
        code=code,
        redirect_uri=redirect_uri,
        verifier=pkce.verifier,
        ssl_ctx=ssl_ctx,
    )

    log.debug("Access token received")
    return token



from nebius.api.buf.validate import validate_pb2 as _validate_pb2
from nebius.api.nebius.common.v1 import metadata_pb2 as _metadata_pb2
from nebius.api.nebius import annotations_pb2 as _annotations_pb2
from google.protobuf.internal import containers as _containers
from google.protobuf.internal import enum_type_wrapper as _enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Iterable as _Iterable, Mapping as _Mapping, Optional as _Optional, Union as _Union

DESCRIPTOR: _descriptor.FileDescriptor

class PoolCidrState(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = []
    STATE_UNSPECIFIED: _ClassVar[PoolCidrState]
    AVAILABLE: _ClassVar[PoolCidrState]
    DISABLED: _ClassVar[PoolCidrState]

class IpVersion(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = []
    IP_VERSION_UNSPECIFIED: _ClassVar[IpVersion]
    IPV4: _ClassVar[IpVersion]
    IPV6: _ClassVar[IpVersion]
STATE_UNSPECIFIED: PoolCidrState
AVAILABLE: PoolCidrState
DISABLED: PoolCidrState
IP_VERSION_UNSPECIFIED: IpVersion
IPV4: IpVersion
IPV6: IpVersion

class Pool(_message.Message):
    __slots__ = ["metadata", "spec", "status"]
    METADATA_FIELD_NUMBER: _ClassVar[int]
    SPEC_FIELD_NUMBER: _ClassVar[int]
    STATUS_FIELD_NUMBER: _ClassVar[int]
    metadata: _metadata_pb2.ResourceMetadata
    spec: PoolSpec
    status: PoolStatus
    def __init__(self, metadata: _Optional[_Union[_metadata_pb2.ResourceMetadata, _Mapping]] = ..., spec: _Optional[_Union[PoolSpec, _Mapping]] = ..., status: _Optional[_Union[PoolStatus, _Mapping]] = ...) -> None: ...

class PoolSpec(_message.Message):
    __slots__ = ["source_pool_id", "source_scope_id", "version", "cidrs"]
    SOURCE_POOL_ID_FIELD_NUMBER: _ClassVar[int]
    SOURCE_SCOPE_ID_FIELD_NUMBER: _ClassVar[int]
    VERSION_FIELD_NUMBER: _ClassVar[int]
    CIDRS_FIELD_NUMBER: _ClassVar[int]
    source_pool_id: str
    source_scope_id: str
    version: IpVersion
    cidrs: _containers.RepeatedCompositeFieldContainer[PoolCidr]
    def __init__(self, source_pool_id: _Optional[str] = ..., source_scope_id: _Optional[str] = ..., version: _Optional[_Union[IpVersion, str]] = ..., cidrs: _Optional[_Iterable[_Union[PoolCidr, _Mapping]]] = ...) -> None: ...

class PoolCidr(_message.Message):
    __slots__ = ["cidr", "state", "allowed_mask"]
    CIDR_FIELD_NUMBER: _ClassVar[int]
    STATE_FIELD_NUMBER: _ClassVar[int]
    ALLOWED_MASK_FIELD_NUMBER: _ClassVar[int]
    cidr: str
    state: PoolCidrState
    allowed_mask: int
    def __init__(self, cidr: _Optional[str] = ..., state: _Optional[_Union[PoolCidrState, str]] = ..., allowed_mask: _Optional[int] = ...) -> None: ...

class PoolStatus(_message.Message):
    __slots__ = ["state", "cidrs", "scope_id"]
    class State(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
        __slots__ = []
        STATE_UNSPECIFIED: _ClassVar[PoolStatus.State]
        CREATING: _ClassVar[PoolStatus.State]
        READY: _ClassVar[PoolStatus.State]
        DELETING: _ClassVar[PoolStatus.State]
    STATE_UNSPECIFIED: PoolStatus.State
    CREATING: PoolStatus.State
    READY: PoolStatus.State
    DELETING: PoolStatus.State
    STATE_FIELD_NUMBER: _ClassVar[int]
    CIDRS_FIELD_NUMBER: _ClassVar[int]
    SCOPE_ID_FIELD_NUMBER: _ClassVar[int]
    state: PoolStatus.State
    cidrs: _containers.RepeatedScalarFieldContainer[str]
    scope_id: str
    def __init__(self, state: _Optional[_Union[PoolStatus.State, str]] = ..., cidrs: _Optional[_Iterable[str]] = ..., scope_id: _Optional[str] = ...) -> None: ...
